"""Plotscripts to visualize circuit execution's result."""

from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from qibo.backends import _check_backend
from qibo.config import raise_error
from qibo.result import CircuitResult, MeasurementOutcomes, QuantumState
from qibo.ui.drawing_utils import (
    QIBO_COMPLEMENTARY_COLOR,
    QIBO_DEFAULT_COLOR,
    generate_bitstring_combinations,
)


def visualize_state(
    execution_outcome: Union[QuantumState, MeasurementOutcomes, CircuitResult],
    mode: str = "probabilities",
    n_most_relevant_components=None,
):
    """
    Plot circuit execution's result data according to the chosen ``mode``.

    Args:
        execution_outcome: qibo circuit's result. Depending on the simulation
            preferences, some of the visualizations can be accessed and some of them
            not. In particular:
                - if ``execution_outcome`` is a `QuantumState`, only probabilities and
                  amplitudes can be visualized;
                - if ``execution_outcome`` is a ``MeasurementOutcomes``, then all
                  the ``mode`` options are available.
        mode: visualization mode can be "amplitudes", "probabilities" or "frequencies".
            Default is "probabilities".
        n_most_relevant_components (int): in case the system is big (more than a few
            qubits), it can be helpful to reduce the number of ticks in the x-axis.
            To do so, this argument can be set, reducing the number of plotted ticks
            to `n_most_relevant_components`. Default is None.
    """
    # Collect amplitude
    amplitudes = execution_outcome.backend.to_numpy(execution_outcome.state())
    nqubits = int(np.log2(len(amplitudes)))

    bitstrings = generate_bitstring_combinations(nqubits)

    num_bitstrings = len(bitstrings)

    fig_width = max(6, nqubits * 1.5)
    fig_height = 5

    _, ax = plt.subplots(figsize=(fig_width, fig_height))

    x = np.arange(num_bitstrings)

    ax.set_xlabel("States")
    ax.set_ylabel(mode.capitalize())

    if mode == "frequencies":
        if not isinstance(execution_outcome, MeasurementOutcomes):
            raise_error(
                ValueError,
                "To visualize frequencies, ensure the circuit is executed with shots.",
            )
        frequencies = execution_outcome.frequencies()
        y_values = [frequencies.get(b, 0) for b in bitstrings]
        ax.bar(x, y_values, color=QIBO_DEFAULT_COLOR, edgecolor="black")

    elif mode == "amplitudes":
        real_parts = np.real(amplitudes)
        imag_parts = np.imag(amplitudes)
        width = 0.3
        ax.bar(
            x - width / 2,
            real_parts,
            width,
            color=QIBO_DEFAULT_COLOR,
            edgecolor="black",
            label="Real part",
        )
        ax.bar(
            x + width / 2,
            imag_parts,
            width,
            color=QIBO_COMPLEMENTARY_COLOR,
            edgecolor="black",
            label="Imag part",
        )
        ax.hlines(0, -1, num_bitstrings, color="black", lw=1)
        ax.legend()
        y_values = np.abs(amplitudes)

    elif mode == "probabilities":
        probabilities = np.abs(amplitudes) ** 2
        ax.bar(x, probabilities, color=QIBO_DEFAULT_COLOR, edgecolor="black")
        y_values = probabilities

    else:
        raise_error(
            ValueError,
            f"Unsupported mode '{mode}'. Choose 'amplitudes', 'probabilities' or 'frequencies'.",
        )

    # Adjust x-axis labels based on number of qubits
    if n_most_relevant_components is not None:
        top_indices = np.argsort(y_values)[-n_most_relevant_components:]
        tick_labels = [
            bitstrings[i] if i in top_indices else "" for i in range(num_bitstrings)
        ]
        ax.set_xticks(x)
        ax.set_xticklabels(tick_labels, rotation=90)
    else:
        ax.set_xticks(x)
        ax.set_xticklabels(bitstrings, rotation=90)

    plt.tight_layout()

    return ax, ax.figure


def _validate_colors(colors):
    if colors is None:
        return QIBO_DEFAULT_COLOR, QIBO_COMPLEMENTARY_COLOR

    if len(colors) != 2:
        raise_error(
            ValueError, f"Colors must be a list of length 2, got {len(colors)} instead"
        )

    pos_color = QIBO_DEFAULT_COLOR if colors[0] is None else colors[0]
    neg_color = QIBO_COMPLEMENTARY_COLOR if colors[1] is None else colors[1]

    return pos_color, neg_color


def _get_labels(nqubits):
    indices = np.arange(2**nqubits)
    bits = np.array(np.unravel_index(indices, [2] * nqubits)).T
    labels = ["".join(str(b) for b in row) for row in bits]
    return labels


# Based on Qiskit density state plot
# https://github.com/Qiskit/qiskit/blob/stable/2.0/qiskit/visualization/state_visualization.py#L372-L622
def plot_density_hist(
    circuit,
    title: str = "",
    alpha: float = 0.5,
    colors: Optional[List[str]] = None,
    fig_width: int = 16,
    fig_height: int = 8,
    backend=None,
    n_most_relevant_components=None,
    **kwargs,
):
    """Plot the real and imaginary parts of the density matrix.

    Given a :class:`qibo.models.circuit.Circuit`, plot the real and imaginary parts
    of the final density matrix as separate 3D cityscape plots, side by side, and
    with a gray ``z=0`` plane for the imaginary part.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Circuit to visualize.
        title (str, optional): Title of the plot. Defaults to ``""``.
        alpha (float, optional): Transparency level for the bars in the plot.
            Defaults to :math:`0.5`.
        colors (list, optional): A list of two colors for the positive and negative
            parts of the density matrix. If ``None``, default colors will be used.
            Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend
            to be used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.
        fig_width (int, optional): Width of the figure in inches. Defaults to ``16``.
        fig_height (int, optional): Height of the figure in inches. Defaults to ``8``.
        n_most_relevant_components (int): in case the system is big (more than a few
            qubits), it can be helpful to reduce the number of ticks in the x-axis.
            To do so, this argument can be set, reducing the number of plotted ticks
            to `n_most_relevant_components`. Default is None.

    Returns:
        tuple: Respectively, the figure, and axes for the real and the imaginary parts.
    """

    # if exec_circ has measurements then, error measure gates are present
    if len(circuit.measurements) > 0:
        raise_error(
            ValueError,
            "Circuit must not contain measurement gates for density matrix visualization",
        )

    backend = _check_backend(backend)
    # Execute the circuit to get the state
    exec_circ = backend.execute_circuit(circuit)

    state = exec_circ.state()

    # Create a density matrix from state vector
    if not circuit.density_matrix:
        state = backend.cast(state)
        state = backend.np.outer(state, backend.np.conj(state))

    nqubits = circuit.nqubits
    row_names = _get_labels(nqubits)
    column_names = _get_labels(nqubits)

    matrix_real = state.real
    matrix_imag = state.imag
    del state

    label_y, label_x = matrix_real.shape[:2]
    xpos = np.arange(0, label_x, 1)
    ypos = np.arange(0, label_y, 1)
    xpos, ypos = np.meshgrid(xpos + 0.25, ypos + 0.25)
    xpos = xpos.flatten()
    ypos = ypos.flatten()
    zpos = np.zeros(label_x * label_y)

    dx = 0.5 * np.ones_like(zpos)
    dy = dx.copy()
    dzr = backend.np.reshape(matrix_real, (-1))
    dzi = backend.np.reshape(matrix_imag, (-1))

    max_dzr = np.max(dzr)
    max_dzi = np.max(dzi)

    fig = plt.figure(figsize=(fig_width, fig_height), facecolor="w")
    ax1 = fig.add_subplot(1, 2, 1, projection="3d", computed_zorder=False)
    ax2 = fig.add_subplot(1, 2, 2, projection="3d", computed_zorder=False)

    max_plot_size = min(fig_width / 2.25, fig_height)
    max_font_size = int(2 * max_plot_size)
    max_zoom = 10 / (10 + np.sqrt(max_plot_size))

    pos_color, neg_color = _validate_colors(colors)
    cmap_pos = LinearSegmentedColormap.from_list("cmap_pos", 4 * [pos_color])
    cmap_neg = LinearSegmentedColormap.from_list("cmap_neg", 4 * [neg_color])

    negative_order = kwargs.get("negative_order", 0.625)
    positive_order = kwargs.get("positive_order", 0.875)
    alpha_3dplane = kwargs.get("alpha_3dplane", 0.25)

    for ax, dz, zlabel in (
        (ax1, dzr, "Real"),
        (ax2, dzi, "Imaginary"),
    ):

        max_dz = np.max(dz)
        min_dz = np.min(dz)

        # Normalize the heights for colormap
        norm_pos = plt.Normalize(vmin=0, vmax=max_dz)
        norm_neg = plt.Normalize(vmin=min_dz, vmax=0)

        # Create a color array based on the heights
        colors_mapping = np.array(
            [
                (
                    cmap_pos(norm_pos(height))
                    if height >= 0
                    else cmap_neg(norm_neg(height))
                )
                for height in dz
            ]
        )

        for mask, zorder in ((dz < 0, negative_order), (dz >= 0, positive_order)):
            if np.any(mask):
                ax.bar3d(
                    xpos[mask],
                    ypos[mask],
                    zpos[mask],
                    dx[mask],
                    dy[mask],
                    dz[mask],
                    alpha=alpha,
                    zorder=zorder,
                    color=colors_mapping[mask],
                    shade=True,
                )

        if min_dz < 0 < max_dz:
            xlim, ylim = [0, label_x], [0, label_y]
            verts = [list(zip(xlim + xlim[::-1], np.repeat(ylim, 2), [0] * 4))]
            plane = Poly3DCollection(
                verts, alpha=alpha_3dplane, facecolor="k", linewidths=1
            )
            plane.set_zorder(0.75)
            ax.add_collection3d(plane)

        ax.set_title(zlabel, fontsize=max_font_size)

        ax.set_xticks(np.arange(0.5, label_x + 0.5, 1))
        ax.set_yticks(np.arange(0.5, label_y + 0.5, 1))

        if max_dz != min_dz:
            ax.axes.set_zlim3d(min_dz, max(max_dzr + 1e-9, max_dzi))
        else:
            if min_dz == 0:
                ax.axes.set_zlim3d(min_dz, max(max_dzr + 1e-9, max_dzi))
            else:
                ax.axes.set_zlim3d(auto=True)
        ax.get_autoscalez_on()

        if n_most_relevant_components is not None:
            total = len(row_names)
            indices = list(range(total))
            if n_most_relevant_components < total:
                step = max(1, total // n_most_relevant_components)
                indices = list(range(0, total, step))[:n_most_relevant_components]
            xtick_labels = [row_names[i] if i in indices else "" for i in range(total)]
            ytick_labels = [
                column_names[i] if i in indices else "" for i in range(total)
            ]
        else:
            xtick_labels = row_names
            ytick_labels = column_names

        ax.xaxis.set_ticklabels(
            xtick_labels, fontsize=max_font_size, rotation=30, ha="right", va="top"
        )
        ax.yaxis.set_ticklabels(
            ytick_labels, fontsize=max_font_size, rotation=30, ha="left", va="center"
        )

        for tick in ax.zaxis.get_major_ticks():
            tick.label1.set_fontsize(max_font_size)
            tick.label1.set_horizontalalignment("left")
            tick.label1.set_verticalalignment("bottom")

        ax.set_xlabel("Ket", labelpad=20)
        ax.set_ylabel("Bra", labelpad=15)
        ax.set_box_aspect(aspect=(4, 4, 4), zoom=max_zoom)
        ax.set_xmargin(0)
        ax.set_ymargin(0)

    if title != "":
        fig.suptitle(title, fontsize=max_font_size)
    fig.subplots_adjust(top=0.9, bottom=0, left=0, right=1, hspace=0, wspace=0)

    return fig, ax1, ax2



# Some functions in MPLDrawer are from code provided by Rick Muller
# Simplified Plotting Routines for Quantum Circuits
# https://github.com/rpmuller/PlotQCircuit
#
import json
from pathlib import Path
from typing import Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.transforms import Bbox

from qibo import gates

from .drawing_utils import FusedEndGateBarrier, FusedStartGateBarrier

UI = Path(__file__).parent
STYLE = json.loads((UI / "styles.json").read_text())
SYMBOLS = json.loads((UI / "symbols.json").read_text())

PLOT_PARAMS = {
    "scale": 1.0,
    "fontsize": 14.0,
    "linewidth": 1.0,
    "control_radius": 0.05,
    "not_radius": 0.15,
    "swap_delta": 0.08,
    "label_buffer": 0.0,
    "dpi": 100,
    "facecolor": "w",
    "edgecolor": "#000000",
    "fillcolor": "#000000",
    "linecolor": "k",
    "textcolor": "k",
    "gatecolor": "w",
    "controlcolor": "#000000",
}


def _plot_quantum_schedule(
    schedule, inits, plot_params, labels=[], plot_labels=True, **kwargs
):
    """Use Matplotlib to plot a queue of quantum circuit.

    Args:
        schedule (list):  List of time steps, each containing a sequence of gates during that step.
        Each gate is a tuple containing (name,target,control1,control2...). Targets and controls initially defined in terms of labels.

        inits (list): Initialization list of gates.

        plot_params (list): Style plot configuration.

        labels (list): List of qubit labels, optional.

        kwargs (list): Variadic list that can override plot parameters.

    Returns:
        matplotlib.axes.Axes: An Axes object encapsulates all the plt elements of a plot in a figure.
    """

    return _plot_quantum_circuit(
        schedule,
        inits,
        plot_params,
        labels=labels,
        plot_labels=plot_labels,
        schedule=True,
        **kwargs
    )


def _plot_quantum_circuit(
    gates, inits, plot_params, labels=[], plot_labels=True, schedule=False, **kwargs
):
    """Use Matplotlib to plot a quantum circuit.

    Args:
        gates (list): List of tuples for each gate in the quantum circuit. (name,target,control1,control2...).
        Targets and controls initially defined in terms of labels.

        inits (list): Initialization list of gates.

        plot_params (list): Style plot configuration.

        labels (list): List of qubit labels. (optional).

        kwargs (list): Variadic list that can override plot parameters.

    Returns:
        matplotlib.axes.Axes: An Axes object encapsulates all the plt elements of a plot in a figure.
    """

    plot_params.update(kwargs)
    scale = plot_params["scale"]

    # Create labels from gates. This will become slow if there are a lot
    #  of gates, in which case move to an ordered dictionary
    if not labels:
        labels = []
        for i, gate in _enumerate_gates(gates, schedule=schedule):
            for label in gate[1:]:
                if label not in labels:
                    labels.append(label)

    nq = len(labels)
    ng = len(gates)

    wire_grid = np.arange(0.0, nq * scale, scale, dtype=float)

    gate_grid = np.arange(0.0, (nq if ng == 0 else ng) * scale, scale, dtype=float)
    ax, _ = _setup_figure(
        nq, (nq if ng == 0 else ng), gate_grid, wire_grid, plot_params
    )

    measured = None if ng == 0 else _measured_wires(gates, labels, schedule=schedule)
    _draw_wires(ax, nq, gate_grid, wire_grid, plot_params, measured)

    if plot_labels:
        _draw_labels(ax, labels, inits, gate_grid, wire_grid, plot_params)

    if ng > 0:
        _draw_gates(
            ax,
            gates,
            labels,
            gate_grid,
            wire_grid,
            plot_params,
            measured,
            schedule=schedule,
        )

    return ax


def _enumerate_gates(gates_plot, schedule=False):
    """Enumerate the gates in a way that can take l as either a list of gates or a schedule

    Args:
        gates_plot (list): List of gates to plot.

        schedule (bool): Check whether process single gate or array of gates at a time.

    Returns:
        int: Index of gate or list of gates.

        list: Processed list of gates ready to plot.
    """

    if schedule:
        for i, gates in enumerate(gates_plot):
            for gate in gates:
                yield i, gate
    else:
        for i, gate in enumerate(gates_plot):
            yield i, gate


def _measured_wires(gates_plot, labels, schedule=False):
    measured = {}
    for i, gate in _enumerate_gates(gates_plot, schedule=schedule):
        name, target = gate[:2]
        j = _get_flipped_index(target, labels)
        if name.startswith("M"):
            measured[j] = i
    return measured


def _draw_gates(
    ax,
    gates_plot,
    labels,
    gate_grid,
    wire_grid,
    plot_params,
    measured={},
    schedule=False,
):
    for i, gate in _enumerate_gates(gates_plot, schedule=schedule):
        _draw_target(ax, i, gate, labels, gate_grid, wire_grid, plot_params)
        if len(gate) > 2:  # Controlled
            _draw_controls(
                ax, i, gate, labels, gate_grid, wire_grid, plot_params, measured
            )


def _draw_controls(ax, i, gate, labels, gate_grid, wire_grid, plot_params, measured={}):

    name, target = gate[:2]

    if "FUSEDENDGATEBARRIER" in name:
        return

    linewidth = plot_params["linewidth"]
    scale = plot_params["scale"]
    control_radius = plot_params["control_radius"]

    target_index = _get_flipped_index(target, labels)
    controls = gate[2:]
    control_indices = _get_flipped_indices(controls, labels)
    gate_indices = control_indices + [target_index]
    min_wire = min(gate_indices)
    max_wire = max(gate_indices)

    if "FUSEDSTARTGATEBARRIER" in name:
        equal_qbits = False
        if "@EQUAL" in name:
            name = name.replace("@EQUAL", "")
            equal_qbits = True
        nfused = int(name.replace("FUSEDSTARTGATEBARRIER", ""))
        dx_right = 0.30
        dx_left = 0.30
        dy = 0.25
        _rectangle(
            ax,
            gate_grid[i + 1] - dx_left,
            gate_grid[i + nfused] + dx_right,
            wire_grid[min_wire] - dy - (0 if not equal_qbits else -0.9 * scale),
            wire_grid[max_wire] + dy,
            plot_params,
        )
    elif not "UNITARY@" in name:
        _line(
            ax,
            gate_grid[i],
            gate_grid[i],
            wire_grid[min_wire],
            wire_grid[max_wire],
            plot_params,
            linestyle=(
                "dashed" if name == "UNITARY" and target.count("_") > 1 else "solid"
            ),
        )

        cci = 0
        for ci in control_indices:
            x = gate_grid[i]
            y = wire_grid[ci]

            is_dagger = False
            if name[-2:] == "DG":
                name = name.replace("DG", "")
                is_dagger = True

            if name == "SWAP":
                _swapx(ax, x, y, plot_params)
            elif name in [
                "ISWAP",
                "SISWAP",
                "FSWAP",
                "FSIM",
                "SYC",
                "GENERALIZEDFSIM",
                "RXX",
                "RYY",
                "RZZ",
                "RZX",
                "RXXYY",
                "G",
                "RBS",
                "ECR",
                "MS",
                "UNITARY",
            ]:

                symbol = SYMBOLS.get(name, name)

                if is_dagger:
                    symbol += r"$\rm{^{\dagger}}$"

                if name == "UNITARY" and target.count("_") > 1:
                    hash_split = controls[cci].split("_")
                    u_gate_single_hash = hash_split[2]
                    global_hash = hash_split[3]
                    index_r = int(hash_split[4])
                    subindex = plot_params["hash_unitary_gates"][index_r][
                        u_gate_single_hash + "-" + global_hash
                    ]
                    symbol = r"$\rm_{{{}}}$".format(subindex) + symbol
                    symbol += r"$\rm_{{{}}}$".format(
                        plot_params["hash_global_unitary_gates"][global_hash]
                    )
                    cci += 1

                _text(ax, x, y, symbol, plot_params, box=True)

            else:
                _cdot(ax, x, y, plot_params)
    else:
        x = gate_grid[min_wire]
        y = wire_grid[len(control_indices)]
        strip_symbol = name.replace("UNITARY@", "")

        if strip_symbol == "":
            strip_symbol = "U_G"

        symbol = r"$\rm{{{}}}$".format(strip_symbol)

        dx_right = 0.45
        dy = 0.25
        _composed_rectangle(
            ax,
            gate_grid[i],
            gate_grid[i] + dx_right,
            wire_grid[min_wire] - dy,
            wire_grid[max_wire] + dy,
            symbol,
            plot_params,
        )


def _draw_target(ax, i, gate, labels, gate_grid, wire_grid, plot_params):
    name, target = gate[:2]

    if (
        "FUSEDSTARTGATEBARRIER" in name
        or "FUSEDENDGATEBARRIER" in name
        or "UNITARY@" in name
    ):
        return

    is_dagger = False
    if name[-2:] == "DG":
        name = name.replace("DG", "")
        is_dagger = True

    symbol = SYMBOLS.get(name, name)  # override name with symbols

    if is_dagger:
        symbol += r"$\rm{^{\dagger}}$"

    x = gate_grid[i]
    target_index = _get_flipped_index(target, labels)
    y = wire_grid[target_index]
    if name in ["CNOT", "TOFFOLI"]:
        _oplus(ax, x, y, plot_params)
    elif name == "SWAP":
        _swapx(ax, x, y, plot_params)
    else:
        if name == "ALIGN":
            symbol = "A({})".format(target[2:])

        if name == "UNITARY" and target.count("_") > 1:
            hash_split = target.split("_")
            hash = hash_split[2]
            global_hash = hash_split[3]
            index_r = int(hash_split[4])
            subindex = plot_params["hash_unitary_gates"][index_r][
                hash + "-" + global_hash
            ]
            symbol = r"$\rm_{{{}}}$".format(subindex) + symbol
            symbol += r"$\rm_{{{}}}$".format(
                plot_params["hash_global_unitary_gates"][global_hash]
            )

        _text(ax, x, y, symbol, plot_params, box=True)


def _line(ax, x1, x2, y1, y2, plot_params, linestyle="solid"):
    Line2D = matplotlib.lines.Line2D
    line = Line2D(
        (x1, x2),
        (y1, y2),
        color=plot_params["linecolor"],
        lw=plot_params["linewidth"],
        ls=linestyle,
    )
    ax.add_line(line)


def _text(ax, x, y, textstr, plot_params, box=False):
    linewidth = plot_params["linewidth"]
    fontsize = (
        12.0
        if _check_list_str(["dagger", "sqrt"], textstr)
        else plot_params["fontsize"]
    )

    if box:
        bbox = dict(
            ec=plot_params["edgecolor"],
            fc=plot_params["gatecolor"],
            fill=True,
            lw=linewidth,
        )
    else:
        bbox = dict(fill=False, lw=0)
    return ax.text(
        x,
        y,
        textstr,
        color=plot_params["textcolor"],
        ha="center",
        va="center",
        bbox=bbox,
        size=fontsize,
    )


def _oplus(ax, x, y, plot_params):
    Line2D = matplotlib.lines.Line2D
    Circle = matplotlib.patches.Circle
    not_radius = plot_params["not_radius"]
    linewidth = plot_params["linewidth"]
    c = Circle(
        (x, y),
        not_radius,
        ec=plot_params["edgecolor"],
        fc=plot_params["gatecolor"],
        fill=True,
        lw=linewidth,
    )
    ax.add_patch(c)
    _line(ax, x, x, y - not_radius, y + not_radius, plot_params)


def _cdot(ax, x, y, plot_params):
    Circle = matplotlib.patches.Circle
    control_radius = plot_params["control_radius"]
    scale = plot_params["scale"]
    linewidth = plot_params["linewidth"]
    c = Circle(
        (x, y),
        control_radius * scale,
        ec=plot_params["edgecolor"],
        fc=plot_params["controlcolor"],
        fill=True,
        lw=linewidth,
    )
    ax.add_patch(c)


def _swapx(ax, x, y, plot_params):
    d = plot_params["swap_delta"]
    linewidth = plot_params["linewidth"]
    _line(ax, x - d, x + d, y - d, y + d, plot_params)
    _line(ax, x - d, x + d, y + d, y - d, plot_params)


def _setup_figure(nq, ng, gate_grid, wire_grid, plot_params):
    scale = plot_params["scale"]
    fig = plt.figure(
        figsize=(ng * scale, nq * scale),
        facecolor=plot_params["facecolor"],
        edgecolor=plot_params["edgecolor"],
        dpi=plot_params["dpi"],
    )
    ax = fig.add_subplot(1, 1, 1, frameon=True)
    ax.set_axis_off()
    offset = 0.5 * scale
    ax.set_xlim(gate_grid[0] - offset, gate_grid[-1] + offset)
    ax.set_ylim(wire_grid[0] - offset, wire_grid[-1] + offset)
    ax.set_aspect("equal")
    return ax, fig


def _draw_wires(ax, nq, gate_grid, wire_grid, plot_params, measured={}):
    scale = plot_params["scale"]
    linewidth = plot_params["linewidth"]
    xdata = (gate_grid[0] - scale, gate_grid[-1] + scale)
    for i in range(nq):
        _line(
            ax,
            gate_grid[0] - scale,
            gate_grid[-1] + scale,
            wire_grid[i],
            wire_grid[i],
            plot_params,
        )


def _draw_labels(ax, labels, inits, gate_grid, wire_grid, plot_params):
    scale = plot_params["scale"]
    label_buffer = plot_params["label_buffer"]
    fontsize = plot_params["fontsize"]
    nq = len(labels)
    xdata = (gate_grid[0] - scale, gate_grid[-1] + scale)
    if "wire_names" in plot_params:
        labels = (
            plot_params["wire_names"] if len(plot_params["wire_names"]) > 0 else labels
        )
    for i in range(nq):
        j = _get_flipped_index(labels[i], labels)
        _text(
            ax,
            xdata[0] - label_buffer,
            wire_grid[j],
            _render_label(labels[i], inits),
            plot_params,
        )


def _get_min_max_qbits(gates):
    def _get_all_tuple_items(iterable):
        t = []
        for each in iterable:
            t.extend(list(each) if isinstance(each, tuple) else [each])
        return tuple(t)

    all_qbits = []
    c_qbits = [t._control_qubits for t in gates.gates]
    t_qbits = [t._target_qubits for t in gates.gates]
    c_qbits = _get_all_tuple_items(c_qbits)
    t_qbits = _get_all_tuple_items(t_qbits)
    all_qbits.append(c_qbits + t_qbits)

    flatten_arr = _get_all_tuple_items(all_qbits)
    return min(flatten_arr), max(flatten_arr)


def _get_flipped_index(target, labels):
    if isinstance(target, str) and target.count("_") > 1:
        end_index = target.find("_" + target.split("_")[2])
        target = target[:end_index]

    nq = len(labels)
    i = labels.index(target)
    return nq - i - 1


def _rectangle(ax, x1, x2, y1, y2, plot_style):
    Rectangle = matplotlib.patches.Rectangle
    x = min(x1, x2)
    y = min(y1, y2)
    w = abs(x2 - x1)
    h = abs(y2 - y1)
    xm = x + w / 2.0
    ym = y + h / 2.0

    rect = Rectangle(
        (x, y),
        w,
        h,
        ec=plot_style["edgecolor"],
        fc=plot_style["fillcolor"],
        fill=False,
        lw=plot_style["linewidth"],
        label="",
    )
    ax.add_patch(rect)


def _composed_rectangle(ax, x1, x2, y1, y2, label, plot_style):
    """
    Draw a rectangle with a label inside.

    Args:
        ax (matplotlib.axes.Axes): Axes object to draw on.
        x1 (float): x-coordinate of the first corner.
        x2 (float): x-coordinate of the second corner.
        y1 (float): y-coordinate of the first corner.
        y2 (float): y-coordinate of the second corner.
        label (str): Label to display inside the rectangle.
        plot_style (dict): Dictionary containing style parameters for the rectangle.
    """
    Rectangle = matplotlib.patches.Rectangle
    x = min(x1, x2)
    y = min(y1, y2)
    w = abs(x2 - x1)
    h = abs(y2 - y1)
    xm = x + w / 2.0
    ym = y + h / 2.0

    rect = Rectangle(
        (x - w / 2.5, y),
        w * 0.9,
        h,
        ec=plot_style["edgecolor"],
        fc=plot_style["gatecolor"],
        lw=plot_style["linewidth"],
        label="",
        fill=True,
        zorder=2,
    )

    ax.add_patch(rect)
    text_gate = _text(ax, x + w * 0.05, y + h / 2, label, plot_style, box=False)
    _auto_fit_fontsize(text_gate, w * 0.8, None, fig=ax.figure, ax=ax)


def _auto_fit_fontsize(text, width, height, fig=None, ax=None):
    """
    Auto-decrease the fontsize of a text object.

    Args:
        text (matplotlib.text.Text)
        width (float): Allowed width in data coordinates.
        height (float): Allowed height in data coordinates.
        fig (matplotlib.figure.Figure): Figure object to use for rendering.
        ax (matplotlib.axes.Axes): Axes object to use for rendering.
    """
    fig = fig or plt.gcf()
    ax = ax or plt.gca()

    # get text bounding box in figure coordinates
    renderer = fig.canvas.get_renderer()
    bbox_text = text.get_window_extent(renderer=renderer)

    # transform bounding box to data coordinates
    bbox_text = Bbox(ax.transData.inverted().transform(bbox_text))

    # evaluate fit and recursively decrease fontsize until text fits
    fits_width = bbox_text.width < width if width else True
    fits_height = bbox_text.height < height if height else True
    if not all((fits_width, fits_height)):
        text.set_fontsize(text.get_fontsize() - 1)
        text.set_weight("bold")
        _auto_fit_fontsize(text, width, height, fig, ax)


def _get_flipped_indices(targets, labels):
    return [_get_flipped_index(t, labels) for t in targets]


def _render_label(label, inits={}):
    if label in inits:
        s = inits[label]
        if s is None:
            return ""
        else:
            return r"$|%s\rangle$" % inits[label]
    return r"$|%s\rangle$" % label


def _check_list_str(substrings, string):
    return any(item in string for item in substrings)


def _make_cluster_gates(gates_items):
    """
    Given a list of gates from a Qibo circuit, this fucntion gathers all gates to reduce the depth of the circuit making the circuit more user-friendly to avoid very large circuits printed on screen.

    Args:
        gates_items (list): List of gates to gather for circuit depth reduction.

    Returns:
        list: List of gathered gates.
    """

    temp_gates = []
    cluster_gates = []

    for item in gates_items:
        if len(item) == 2:  # single qubit gates
            if len(temp_gates) > 0:
                if item[1] in [tup[1] for tup in temp_gates]:
                    cluster_gates.append(temp_gates)
                    temp_gates = []
                    temp_gates.append(item)
                else:
                    temp_gates.append(item)
            else:
                temp_gates.append(item)
        else:
            if len(temp_gates) > 0:
                cluster_gates.append(temp_gates)
                temp_gates = []

            cluster_gates.append([item])

    if len(temp_gates) > 0:
        cluster_gates.append(temp_gates)

    return cluster_gates


def _build_hash_init_unitary_gates_register(gates_circ, dict_init_param):
    """
    Given a list of gates, this function builds a dictionary to register a unique hash for each unitary gate.

    Args:
        gates_circ (list): List of gates provided by the Qibo circuit.
        dict_init_param (dict): Dictionary to store the hash from unitary gates.
    """
    i = 0
    for gate in gates_circ:
        final_hash = _global_gate_hash(gate)
        if final_hash != "" and final_hash not in dict_init_param:
            dict_init_param[final_hash] = str(i)
            i += 1
    return dict_init_param


def _build_unitary_gates_register(gate, array_register):
    """
    Given a gate, this function builds a dictionary to register the unitary gates and their parameters to identify them uniquely. Only for Unitary gates.

    Args:
        gate (gates.Unitary): Unitary gate to register.
        dict_register (dict): Dictionary to store the unitary gates and their parameters.
    """
    if (
        isinstance(gate, gates.Unitary)
        and len(gate._target_qubits) > 1
        and gate.name.upper() == "UNITARY"
    ):
        i = 0
        dict_register = {}
        for qbit in gate._target_qubits:
            final_hash = _u_hash(gate, qbit)
            param_init_hash = _global_gate_hash(gate)
            dict_register[final_hash + "-" + param_init_hash] = str(i)
            i += 1
        array_register.append(dict_register)


def _global_gate_hash(gate: gates.Unitary):
    """
    Given a unitary gate, this function returns a hash to identify the gate uniquely. Only for Unitary gates.

    Args:
        gate (gates.Unitary): Unitary gate.
    """
    if (
        isinstance(gate, gates.Unitary)
        and len(gate._target_qubits) > 1
        and gate.name.upper() == "UNITARY"
    ):
        hash_result = str(abs(hash(gate._parameters[0].data.tobytes())))
        hash_result = hash_result[:3] + hash_result[-3:]
        return hash_result
    return ""


def _u_hash(gate: gates.Unitary, param_index: int):
    """
    Given a unitary gate and a qubit, this function returns a hash to identify the gate uniquely. Only for Unitary gates.

    Args:
        gate (gates.Unitary): Unitary gate.
        param_index (int): Parameter index applied.
    """
    hash_result = str(abs(hash(gate._parameters[0][param_index].data.tobytes())))
    hash_result = hash_result[:3] + hash_result[-3:]
    return hash_result


def _process_gates(array_gates, nqubits):
    """
    Transforms the list of gates given by the Qibo circuit into a list of gates with a suitable structre to print on screen with matplotlib.

    Args:
        array_gates (list): List of gates provided by the Qibo circuit.
        nqubits (int): Number of circuit qubits

    Returns:
        list: List of suitable gates to plot with matplotlib.
    """

    if len(array_gates) == 0:
        return []

    gates_plot = []
    ucount = 0
    for gate in array_gates:
        init_label = gate.name.upper()

        if init_label == "CCX":
            init_label = "TOFFOLI"
        elif init_label == "CX":
            init_label = "CNOT"
        elif _check_list_str(["SX", "CSX"], init_label):
            is_dagger = init_label[-2:] == "DG"
            init_label = (
                r"$\rm{\sqrt{X}}^{\dagger}$" if is_dagger else r"$\rm{\sqrt{X}}$"
            )
        elif (
            len(gate._control_qubits) > 0
            and "C" in init_label[0]
            and "CNOT" != init_label
        ):
            init_label = gate.draw_label.upper()

        if init_label in [
            "ID",
            "MEASURE",
            "KRAUSCHANNEL",
            "UNITARYCHANNEL",
            "DEPOLARIZINGCHANNEL",
            "READOUTERRORCHANNEL",
        ]:
            for qbit in gate._target_qubits:
                item = (init_label,)
                qbit_item = qbit if qbit < nqubits else nqubits - 1
                item += ("q_" + str(qbit_item),)
                gates_plot.append(item)
        elif init_label == "ENTANGLEMENTENTROPY":
            for qbit in list(range(nqubits)):
                item = (init_label,)
                qbit_item = qbit if qbit < nqubits else nqubits - 1
                item += ("q_" + str(qbit_item),)
                gates_plot.append(item)
        else:
            item = ()
            if (
                isinstance(gate, gates.Unitary)
                and len(gate._target_qubits) > 1
                and init_label != "UNITARY"
            ):
                item += ("UNITARY@" + gate.name,)
            else:
                item += (init_label,)

            for qbit in gate._target_qubits:
                if type(qbit) is tuple:
                    qbit_item = qbit[0] if qbit[0] < nqubits else nqubits - 1
                    item += ("q_" + str(qbit_item),)
                else:
                    qbit_item = qbit if qbit < nqubits else nqubits - 1
                    u_param_hash = ""
                    u_global_hash = ""
                    if (
                        isinstance(gate, gates.Unitary)
                        and len(gate._target_qubits) > 1
                        and init_label == "UNITARY"
                    ):
                        u_param_hash = _u_hash(gate, qbit_item)
                        u_global_hash = _global_gate_hash(gate)

                    item += (
                        "q_"
                        + str(qbit_item)
                        + ("" if u_param_hash == "" else ("_" + u_param_hash))
                        + ("" if u_global_hash == "" else ("_" + u_global_hash))
                        + (
                            ("_" + str(ucount))
                            if isinstance(gate, gates.Unitary)
                            and len(gate._target_qubits) > 1
                            and init_label == "UNITARY"
                            else ""
                        ),
                    )

            if (
                isinstance(gate, gates.Unitary)
                and len(gate._target_qubits) > 1
                and init_label == "UNITARY"
            ):
                ucount += 1

            for qbit in gate._control_qubits:
                if type(qbit) is tuple:
                    qbit_item = qbit[0] if qbit[0] < nqubits else nqubits - 1
                    item += ("q_" + str(qbit_item),)
                else:
                    qbit_item = qbit if qbit < nqubits else nqubits - 1
                    item += ("q_" + str(qbit_item),)

            gates_plot.append(item)

    return gates_plot


def _plot_params(style: Union[dict, str, None]) -> dict:
    """
    Given a style name, the function gets the style configuration, if the style is not available, it return the default style. It is allowed to give a custom dictionary to give the circuit a style.

    Args:
        style (Union[dict, str, None]): Name of the style.

    Returns:
        dict: Style configuration.
    """
    if not isinstance(style, dict):
        style = (
            STYLE.get(style)
            if (style is not None and style in STYLE.keys())
            else STYLE["default"]
        )

    return style


def plot_circuit(circuit, scale=0.6, cluster_gates=True, style=None):
    """Main matplotlib plot function for Qibo circuit

    Args:
        circuit (qibo.models.circuit.Circuit): A Qibo circuit to plot.
        scale (float): Scaling factor for matplotlib output drawing.
        cluster_gates (boolean): Group (or not) circuit gates on drawing.
        style (Union[dict, str, None]): Style applied to the circuit, it can a built-in sytle or custom
        (built-in styles: garnacha, fardelejo, quantumspain, color-blind, cachirulo or custom dictionary).

    Returns:
        matplotlib.axes.Axes: Axes object that encapsulates all the elements of an individual plot in a figure.

        matplotlib.figure.Figure: A matplotlib figure object.

    Example:

        .. testcode::

            import matplotlib.pyplot as plt
            import qibo
            from qibo import gates, models
            from qibo.models import QFT

            # new plot function based on matplotlib
            from qibo.ui import plot_circuit

            %matplotlib inline

            # create a 5-qubits QFT circuit
            c = QFT(5)
            c.add(gates.M(qubit) for qubit in range(2))

            # print circuit with default options (default black & white style, scale factor of 0.6 and clustered gates)
            plot_circuit(c);

            # print the circuit with built-int style "garnacha", clustering gates and a custom scale factor
            # built-in styles: "garnacha", "fardelejo", "quantumspain", "color-blind", "cachirulo" or custom dictionary
            plot_circuit(c, scale = 0.8, cluster_gates = True, style="garnacha");

            # plot the Qibo circuit with a custom style
            custom_style = {
                "facecolor" : "#6497bf",
                "edgecolor" : "#01016f",
                "linecolor" : "#01016f",
                "textcolor" : "#01016f",
                "fillcolor" : "#ffb9b9",
                "gatecolor" : "#d8031c",
                "controlcolor" : "#360000"
            }

            plot_circuit(c, scale = 0.8, cluster_gates = True, style=custom_style);
    """

    params = PLOT_PARAMS.copy()
    params.update(_plot_params(style))

    inits = list(range(circuit.nqubits))

    labels = []
    for i in range(circuit.nqubits):
        labels.append("q_" + str(i))

    hash_unitary_gates = []
    all_gates = []
    for gate in circuit.queue:

        _build_unitary_gates_register(gate, hash_unitary_gates)

        if isinstance(gate, gates.FusedGate):
            min_q, max_q = _get_min_max_qbits(gate)

            fgates = None

            if cluster_gates and circuit.nqubits > 1:
                fgates = _make_cluster_gates(
                    _process_gates(gate.gates, circuit.nqubits)
                )
            else:
                fgates = _process_gates(gate.gates, circuit.nqubits)

            l_gates = len(gate.gates)
            equal_qbits = False
            if min_q != max_q:
                l_gates = len(fgates)
            else:
                max_q += 1
                equal_qbits = True

            all_gates.append(FusedStartGateBarrier(min_q, max_q, l_gates, equal_qbits))
            all_gates += gate.gates
            all_gates.append(FusedEndGateBarrier(min_q, max_q))
        else:
            all_gates.append(gate)

    gates_plot = _process_gates(all_gates, circuit.nqubits)

    if hash_unitary_gates:
        params["hash_unitary_gates"] = hash_unitary_gates

    hash_global_unitary_gates = {}
    _build_hash_init_unitary_gates_register(circuit.queue, hash_global_unitary_gates)

    if hash_global_unitary_gates:
        params["hash_global_unitary_gates"] = hash_global_unitary_gates

    params["wire_names"] = (
        circuit.init_kwargs["wire_names"]
        if circuit.init_kwargs["wire_names"] is not None
        else []
    )

    if cluster_gates and len(gates_plot) > 0 and circuit.nqubits > 1:
        gates_cluster = _make_cluster_gates(gates_plot)
        ax = _plot_quantum_schedule(gates_cluster, inits, params, labels, scale=scale)
        return ax, ax.figure

    ax = _plot_quantum_circuit(gates_plot, inits, params, labels, scale=scale)
    return ax, ax.figure



"""Module defining the `QuantumNetwork` class and adjacent functions."""

from functools import reduce
from logging import warning
from operator import mul
from typing import List, Optional, Tuple, Union

import numpy as np

from qibo.backends import _check_backend
from qibo.config import raise_error


class QuantumNetwork:
    """This class stores the representation of the quantum network as a tensor.
    This is a unique representation of the quantum network.

    A minimum quantum network is a quantum channel, which is a quantum network of the form
    :math:`J[n \\to m]`, where :math:`n` is the dimension of the input system ,
    and :math:`m` is the dimension of the output system.
    A quantum state is a quantum network of the form :math:`J: 1 \\to n`,
    such that the input system is trivial.
    An observable is a quantum network of the form :math:`J: n \\to 1`,
    such that the output system is trivial.

    A quantum network may contain multiple input and output systems.
    For example, a "quantum comb" is a quantum network of the form :math:`J: n', n \\to m, m'`,
    which convert a quantum channel of the form :math:`J: n \\to m`
    to a quantum channel of the form :math:`J: n' \\to m'`.

    Args:
        tensor (ndarray): input Choi operator.
        partition (List[int] or Tuple[int]): partition of ``tensor``.
        system_input (List[bool] or Tuple[bool], optional): mask on the output system of the
            Choi operator. If ``None``, defaults to
            ``(True,False,True,False,...)``, where ``len(system_input)=len(partition)``.
            Defaults to ``None``.
        pure (bool, optional): ``True`` when ``tensor`` is a "pure" representation (e.g. a pure
            state, a unitary operator, etc.), ``False`` otherwise. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to the current backend.
            Defaults to ``None``.
    """

    def __init__(
        self,
        tensor,
        partition: Optional[Union[List[int], Tuple[int]]] = None,
        system_input: Optional[Union[List[bool], Tuple[bool]]] = None,
        pure: bool = False,
        backend=None,
    ):
        self._tensor = tensor
        self.partition = tuple(partition)
        self.system_input = system_input
        self._pure = pure
        self._backend = backend

        self._run_checks(self.partition, self.system_input, self._pure)

        self._set_parameters()

        self.dims = reduce(mul, self.partition) if len(self.partition) > 0 else 1

    @staticmethod
    def _order_tensor_to_operator(dims: int):
        """Returns the order to reshape a tensor into an operator.

        Given a tenosr of ``2 * dims`` leads, the order is
        :math:`[0, 2, 4, ..., 1, 3, 5, ...]`.

        Args:
            dims (int): dimension.

        Returns:
            list: order to reshape tensor into an operator.
        """
        return list(range(0, 2 * dims, 2)) + list(range(1, 2 * dims, 2))

    @staticmethod
    def _order_operator_to_tensor(nsystems: int):
        """Returns the order to reshape an operator to a tensor.

        Given a operator of :math:`2n` systems, the order is
        :math:`[0, n, 1, n+1, 2, n+2, ...]`.

        Args:
            nsystems (int): number of systems.

        Returns:
            list: order to reshape operator into tensor.
        """
        return list(
            sum(zip(list(range(0, nsystems)), list(range(nsystems, nsystems * 2))), ())
        )

    @classmethod
    def _operator_to_tensor(cls, operator, partition: List[int]):

        n = len(partition)
        order = cls._order_operator_to_tensor(n)

        # Check if the `partition` matches the shape of the input matrix
        if np.prod(tuple(operator.shape)) != np.prod(
            tuple(dim**2 for dim in partition)
        ):
            raise_error(
                ValueError,
                "``partition`` does not match the shape of the input matrix. "
                + f"Cannot reshape matrix of size {operator.shape} to partition {partition}",
            )

        # Check if `operator` is a pytourch tensor
        tensor = operator.reshape(list(partition) * 2)
        if operator.__class__.__name__ == "Tensor":
            tensor = tensor.permute(order)
        else:
            tensor = tensor.transpose(order)
        return tensor.reshape([dim**2 for dim in partition])

    @classmethod
    def from_operator(
        cls,
        operator,
        partition: Optional[Union[List[int], Tuple[int]]] = None,
        system_input: Optional[Union[List[bool], Tuple[bool]]] = None,
        pure: bool = False,
        backend=None,
    ):
        """Construct a :class:`qibo.quantum_info.QuantumNetwork` object from a ndarray.

        This method converts a Choi operator to the internal representation of
        :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`.
        The input array can be a pure state, a Choi operator, a unitary operator, etc.

        Args:
            arr (ndarray): input numpy array.
            partition (List[int] or Tuple[int], optional): partition of ``arr``. If ``None``,
                defaults to the shape of ``arr``. Defaults to ``None``.
            system_input (List[bool] or Tuple[bool], optional): mask on the input system of the
                Choi operator. If ``None``, defaults to
                ``(True,False,True,False...)``, where ``len(system_input)=len(partition)``.
                Defaults to ``None``.
            pure (bool, optional): ``True`` when ``arr`` is a "pure" representation (e.g. a pure
                state, a unitary operator, etc.), ``False`` otherwise. Defaults to ``False``.
            backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
                calculations. If ``None``, defaults to the current backend.
                Defaults to ``None``.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`:
                quantum network constructed from the input Choi operator.
        """

        if pure:
            if partition is None:
                partition = tuple(operator.shape)
                tensor = operator
            else:
                if tuple(partition) not in [
                    tuple(operator.shape),
                    tuple(int(np.sqrt(dim)) for dim in operator.shape) * 2,
                ]:
                    raise_error(
                        ValueError,
                        "``partition`` does not match the shape of the input matrix. "
                        + f"Cannot reshape matrix of size {operator.shape} "
                        + f"to partition {partition}",
                    )

                tensor = operator.reshape(partition)
        else:
            # check if arr is a valid choi operator
            len_sys = len(operator.shape)
            if (len_sys % 2 != 0) or (
                operator.shape[: len_sys // 2] != operator.shape[len_sys // 2 :]
            ):
                raise_error(
                    ValueError,
                    "The opertor must be a square operator where the first half of the shape "
                    + "is the same as the second half of the shape. "
                    + f"However, the shape of the input is {operator.shape}. "
                    + "If the input is pure, set `pure=True`.",
                )

            if partition is None:
                partition = operator.shape[: len_sys // 2]

            tensor = cls._operator_to_tensor(operator, partition)

        return cls(
            tensor,
            partition=partition,
            system_input=system_input,
            pure=pure,
            backend=backend,
        )

    def operator(self, full: bool = False, backend=None):
        """Returns the Choi operator of the quantum network.

        The shape of the returned operator is :math:`(*self.partition, *self.partition)`.

        Args:
            full (bool, optional): If this is ``False``, and the network is pure, the method
                will only return the eigenvector (unique when the network is pure).
                If ``True``, returns the full tensor of the quantum network. Defaults to ``False``.
            backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used
                to return the Choi operator. If ``None``, defaults to the backend defined
                when initializing the :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`
                object. Defaults to ``None``.

        Returns:
            ndarray: Choi operator of the quantum network.
        """
        if backend is None:  # pragma: no cover
            backend = self._backend

        if self.is_pure() and not full:
            return backend.cast(self._tensor, dtype=self._tensor.dtype)

        tensor = self.full(backend) if self.is_pure() else self._tensor

        n = len(self.partition)
        order = self._order_tensor_to_operator(n)

        operator = self._backend.np.transpose(
            tensor.reshape(tuple(np.repeat(self.partition, 2))), order
        )

        return backend.cast(operator, dtype=self._tensor.dtype)

    def matrix(self, backend=None):
        """Returns the Choi operator of the quantum network in the matrix form.
        The shape of the returned operator is :math:`(self.dims, self.dims)`.

        Args:
            backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used
                to return the Choi operator. If ``None``, defaults to the backend defined
                when initializing the :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`
                object. Defaults to ``None``.

        Returns:
            ndarray: Choi operator of the quantum network.
        """
        return self.operator(full=True, backend=backend).reshape((self.dims, self.dims))

    def is_pure(self):
        """Returns bool indicading if the Choi operator of the network is pure."""
        return self._pure

    def is_hermitian(
        self, order: Optional[Union[int, str]] = None, precision_tol: float = 1e-8
    ):
        """Returns bool indicating if the Choi operator :math:`\\mathcal{J}` is Hermitian.

        Hermicity is calculated as distance between :math:`\\mathcal{J}` and
        :math:`\\mathcal{J}^{\\dagger}` with respect to a given norm.
        Default is the ``Hilbert-Schmidt`` norm (also known as ``Frobenius`` norm).

        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm
        <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.

        Args:
            order (str or int, optional): order of the norm. Defaults to ``None``.
            precision_tol (float, optional): threshold :math:`\\epsilon` that defines if
                Choi operator of the network is :math:`\\epsilon`-close to Hermicity in
                the norm given by ``order``. Defaults to :math:`10^{-8}`.

        Returns:
            bool: Hermiticity condition. If the adjoint of the Choi operator is equal to the
                Choi operator, the method returns ``True``.
                If the input is pure, the its always Hermitian.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if order is None and self._backend.__class__.__name__ == "TensorflowBackend":
            order = "euclidean"

        if self.is_pure():  # if the input is pure, it is always hermitian
            return True

        reshaped = self._backend.cast(
            self.matrix(),
            dtype=self._tensor.dtype,
        )
        if self._backend.__class__.__name__ == "PyTorchBackend":
            adjoint = self._backend.np.transpose(reshaped, (1, 0))
        else:
            adjoint = self._backend.np.transpose(reshaped)

        mat_diff = self._backend.np.conj(adjoint) - reshaped
        norm = self._backend.calculate_matrix_norm(mat_diff, order=order)

        return float(norm) <= precision_tol

    def is_positive_semidefinite(self, precision_tol: float = 1e-8):
        """Returns bool indicating if Choi operator :math:`\\mathcal{J}` is positive-semidefinite.

        Args:
            precision_tol (float, optional): threshold value used to check if eigenvalues of
                the Choi operator :math:`\\mathcal{J}` are such that
                :math:`\\textup{eigenvalues}(\\mathcal{J}) >= - \\textup{precision_tol}`.
                Note that this parameter can be set to negative values.
                Defaults to :math:`0.0`.

        Returns:
            bool: Positive-semidefinite condition.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if self.is_pure():  # if the input is pure, it is always positive semidefinite
            return True

        reshaped = self._backend.cast(
            self.matrix(),
            dtype=self._tensor.dtype,
        )

        if self.is_hermitian():
            eigenvalues = self._backend.calculate_eigenvalues(reshaped)
        else:
            return False

        return all(
            self._backend.np.real(eigenvalue) >= -precision_tol
            for eigenvalue in eigenvalues
        )

    def link_product(self, subscripts: str, second_network):
        """Link product between two quantum networks.

        The link product is not commutative. Here, we assume that
        :math:`A.\\textup{link_product}(B)` means "applying :math:`B` to :math:`A`".
        However, the ``link_product`` is associative, so we override the `@` operation
        in order to simplify notation.

        Args:
            subscripts (str, optional): Specifies the subscript for summation using
                the Einstein summation convention. For more details, please refer to
                `numpy.einsum
                <https://numpy.org/doc/stable/reference/generated/numpy.einsum.html>`_.
            second_network (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum
                network to be applied to the original network.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network resulting
                from the link product between two quantum networks.
        """

        return link_product(subscripts, self, second_network, backend=self._backend)

    def copy(self):
        """Returns a copy of the :class:`qibo.quantum_info.QuantumNetwork` object."""
        return self.__class__(
            self._backend.np.copy(self._tensor),
            partition=self.partition,
            system_input=self.system_input,
            pure=self._pure,
            backend=self._backend,
        )

    def conj(self):
        """Returns the conjugate of the quantum network."""
        return self.__class__(
            self._backend.np.conj(self._tensor),
            partition=self.partition,
            system_input=self.system_input,
            pure=self._pure,
            backend=self._backend,
        )

    def __add__(self, second_network):
        """Add two Quantum Networks by adding their Choi operators.

        This operation always returns a non-pure Quantum Network.

        Args:
            second_network (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum
                network to be added to the original network.

        Returns:
            (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum network resulting
                from the summation of two Choi operators.
        """
        if not isinstance(second_network, QuantumNetwork):
            raise_error(
                TypeError,
                "It is not possible to add a object of type ``QuantumNetwork`` "
                + f"and and object of type ``{type(second_network)}``.",
            )

        if self.full().shape != second_network.full().shape:
            raise_error(
                ValueError,
                f"The Choi operators must have the same shape, but {self.full().shape} != "
                + f"{second_network.full().shape}.",
            )

        if self.system_input != second_network.system_input:
            raise_error(ValueError, "The networks must have the same input systems.")

        new_first_tensor = self.full()
        new_second_tensor = second_network.full()

        return QuantumNetwork(
            new_first_tensor + new_second_tensor,
            self.partition,
            self.system_input,
            pure=False,
            backend=self._backend,
        )

    def __mul__(self, number: Union[float, int]):
        """Returns quantum network with its Choi operator multiplied by a scalar.

        If the quantum network is pure and ``number > 0.0``, the method returns a pure quantum
        network with its Choi operator multiplied by the square root of ``number``.
        This is equivalent to multiplying `self.to_full()` by the ``number``.
        Otherwise, this method will return a full quantum network.

        Args:
            number (float or int): scalar to multiply the Choi operator of the network with.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network with its
                Choi operator multiplied by ``number``.
        """
        if not isinstance(number, (float, int)):
            raise_error(
                TypeError,
                "It is not possible to multiply a ``QuantumNetwork`` by a non-scalar.",
            )

        if self.is_pure() and number > 0.0:
            return QuantumNetwork(
                np.sqrt(number) * self._tensor,
                partition=self.partition,
                system_input=self.system_input,
                pure=True,
                backend=self._backend,
            )

        tensor = self.full()

        return QuantumNetwork(
            number * tensor,
            partition=self.partition,
            system_input=self.system_input,
            pure=False,
            backend=self._backend,
        )

    def __rmul__(self, number: Union[float, int]):
        """"""
        return self.__mul__(number)

    def __truediv__(self, number: Union[float, int]):
        """Returns quantum network with its Choi operator divided by a scalar.

        If the quantum network is pure and ``number > 0.0``, the method returns a pure quantum
        network with its Choi operator divided by the square root of ``number``.
        This is equivalent to dividing `self.to_full()` by the ``number``.
        Otherwise, this method will return a full quantum network.

        Args:
            number (float or int): scalar to divide the Choi operator of the network with.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network with its
                Choi operator divided by ``number``.
        """
        if not isinstance(number, (float, int)):
            raise_error(
                TypeError,
                "It is not possible to divide a ``QuantumNetwork`` by a non-scalar.",
            )

        number = np.sqrt(number) if self.is_pure() and number > 0.0 else number

        return QuantumNetwork(
            self._tensor / number,
            partition=self.partition,
            system_input=self.system_input,
            pure=self.is_pure(),
            backend=self._backend,
        )

    def __matmul__(self, second_network):
        """Defines matrix multiplication between two ``QuantumNetwork`` objects.

        If ``self.partition == second_network.partition in [2, 4]``, this method is overwritten by
        :meth:`qibo.quantum_info.quantum_networks.QuantumNetwork.link_product`.

        Args:
            second_network (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`):

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network resulting
                from the link product operation.
        """
        if not isinstance(second_network, QuantumNetwork):
            raise_error(
                TypeError,
                "It is not possible to implement matrix multiplication of a "
                + "``QuantumNetwork`` by a non-``QuantumNetwork``.",
            )

        if len(self.partition) == 2:  # `self` is a channel
            if len(second_network.partition) != 2:
                raise_error(
                    ValueError,
                    f"`QuantumNetwork {second_network} is assumed to be a channel, but it is not. "
                    + "Use `link_product` method to specify the subscript.",
                )
            if self.partition[1] != second_network.partition[0]:
                raise_error(
                    ValueError,
                    "partitions of the networks do not match: "
                    + f"{self.partition[1]} != {second_network.partition[0]}.",
                )

            subscripts = "jk,kl -> jl"

        elif len(self.partition) == 4:  # `self` is a super-channel
            if len(second_network.partition) != 2:
                raise_error(
                    ValueError,
                    f"`QuantumNetwork {second_network} is assumed to be a channel, but it is not. "
                    + "Use `link_product` method to specify the subscript.",
                )
            if self.partition[1] != second_network.partition[0]:
                raise_error(
                    ValueError,
                    "Systems of the channel do not match the super-channel: "
                    + f"{self.partition[1], self.partition[2]} != "
                    + f"{second_network.partition[0],second_network.partition[1]}.",
                )

            subscripts = "jklm,kl -> jm"
        else:
            raise_error(
                NotImplementedError,
                "`partitions` do not match any implemented pattern``. "
                + "Use `link_product` method to specify the subscript.",
            )

        return self.link_product(subscripts, second_network)  # pylint: disable=E0606

    def __str__(self):
        """Method to define how to print relevant information of the quantum network."""
        systems = []

        for i, dim in enumerate(self.partition):
            if self.system_input[i]:
                systems.append(f"┍{dim}┑")
            else:
                systems.append(f"┕{dim}┙")

        return f"J[{', '.join(systems)}]"

    def _run_checks(self, partition, system_input, pure):
        """Checks if all inputs are correct in type and value."""
        if not isinstance(partition, (list, tuple)):
            raise_error(
                TypeError,
                "``partition`` must be type ``tuple`` or ``list``, "
                + f"but it is type ``{type(partition)}``.",
            )

        if any(not isinstance(party, int) for party in partition):
            raise_error(
                ValueError,
                "``partition`` must be a ``tuple`` or ``list`` of positive integers, "
                + "but contains non-integers.",
            )

        if any(party <= 0 for party in partition):
            raise_error(
                ValueError,
                "``partition`` must be a ``tuple`` or ``list`` of positive integers, "
                + "but contains non-positive integers.",
            )

        if system_input is not None and len(system_input) != len(partition):
            raise_error(
                ValueError,
                "``len(system_input)`` must be the same as ``len(partition)``, "
                + f"but {len(system_input)} != {len(partition)}.",
            )

        if not isinstance(pure, bool):
            raise_error(
                TypeError,
                f"``pure`` must be type ``bool``, but it is type ``{type(pure)}``.",
            )

    @staticmethod
    def _check_system_input(system_input, partition) -> Tuple[bool]:
        """
        If `system_input` not defined, assume the network follows the order of a quantum Comb.
        """

        if system_input is None:
            system_input = [
                False,
            ] * len(partition)
            for k in range(len(partition) // 2):
                system_input[k * 2] = True
        return tuple(system_input)

    def _set_parameters(self):
        """Standarize the parameters."""
        self._backend = _check_backend(self._backend)

        self.partition = tuple(self.partition)

        self.system_input = self._check_system_input(self.system_input, self.partition)

        self._einsum = self._backend.np.einsum
        self._tensordot = self._backend.np.tensordot
        self._tensor = self._backend.cast(self._tensor, dtype=self._tensor.dtype)

        if self._pure:
            if np.prod(tuple(self._tensor.shape)) != np.prod(tuple(self.partition)):
                raise_error(
                    ValueError,
                    "``partition`` does not match the shape of the input matrix. "
                    + f"Cannot reshape matrix of size {self._tensor.shape} "
                    + f"to partition {self.partition}.",
                )
            self._tensor = self._backend.np.reshape(self._tensor, self.partition)
        else:
            if np.prod(tuple(self._tensor.shape)) != np.prod(
                tuple(dim**2 for dim in self.partition)
            ):
                raise_error(
                    ValueError,
                    "``partition`` does not match the shape of the input matrix. "
                    + f"Cannot reshape matrix of size {self._tensor.shape} "
                    + f"to partition {self.partition}.",
                )
            matrix_partition = [dim**2 for dim in self.partition]
            self._tensor = self._backend.np.reshape(self._tensor, matrix_partition)

    def full(self, update: bool = False, backend=None):
        """Convert the internal representation to the full tensor of the network.

        Args:
            update (bool, optional): If ``True``, updates the internal representation of the
                network to the full tensor. Defaults to ``False``.
            backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
                calculations. If ``None``, defaults to the current backend.
                Defaults to ``None``.

        Returns:
            ndarray: full reprentation of the quantum network.
        """
        if backend is None:  # pragma: no cover
            backend = self._backend
        tensor = self._backend.np.copy(self._tensor)
        tensor = backend.cast(tensor, dtype=self._tensor.dtype)
        conj = backend.np.conj

        if self.is_pure():
            # Reshapes input matrix based on purity.
            tensor.reshape(self.dims)
            if self._backend.__class__.__name__ == "PyTorchBackend":
                tensor = self._tensordot(tensor, conj(tensor), dims=0)
            else:
                tensor = self._tensordot(tensor, conj(tensor), axes=0)
            tensor = self._operator_to_tensor(tensor, self.partition)

            if update:
                self._tensor = tensor
                self._pure = False

        return tensor


class QuantumComb(QuantumNetwork):
    """Stores a Quantum comb, which is a network in which the systems follows a sequential order.

    It is also called the *non-Markovian quantum process* in many literatures.
    A quantum comb is a quantum network of the form :math:`J[┍i1┑,┕o1┙,┍i2┑,┕o2┙, ...]`,
    where the process first take an input state from system :math:`i1`,
    then output a state to system :math:`o1`, and so on.
    This is a non-Markovian process as the output of the system :math:`o2` may depend on
    what happened in systems :math:`i1`, and :math:`o1`.

    A quantum channel is a special case of quantum comb, where there are only one input
    system and one output system.

    Args:
        tensor (ndarray): the tensor representations of the quantum Comb.
        partition (List[int] or Tuple[int]): partition of ``matrix``.
        system_input (List[bool] or Tuple[bool], optional): mask on the input system of the
            Choi operator. If ``None``, defaults to
            ``(True,False,True,False,...)``, where ``len(system_input)=len(partition)``.
            Defaults to ``None``.
        pure (bool, optional): ``True`` when ``tensor`` is a "pure" representation (e.g. a pure
            state, a unitary operator, etc.), ``False`` otherwise. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to the current backend.
            Defaults to ``None``.
    """

    def __init__(
        self,
        tensor,
        partition: Optional[Union[List[int], Tuple[int]]] = None,
        system_input: Optional[Union[List[bool], Tuple[bool]]] = None,
        pure: bool = False,
        backend=None,
    ):
        if partition is None:
            if pure:
                partition = tensor.shape
            else:
                partition = tuple(int(np.sqrt(d)) for d in tensor.shape)
        if len(partition) % 2 != 0:
            raise_error(
                ValueError,
                "A quantum comb should only contain equal number of input and output systems. "
                + "For general quantum networks, one should use the ``QuantumNetwork`` class.",
            )
        if system_input is not None:
            warning("system_input is ignored for QuantumComb")

        super().__init__(
            tensor, partition, [True, False] * (len(partition) // 2), pure, backend
        )

    def is_causal(
        self, order: Optional[Union[int, str]] = None, precision_tol: float = 1e-8
    ):
        """Returns bool indicating if the Choi operator :math:`\\mathcal{J}` satisfies causal order

        Causality is calculated based on a recursive constrains.
        This method reduce a n-comb to a (n-1)-comb at each step,
        and checks if the reduced comb is independent on the last output system.

        Args:
            order (str or int, optional): order of the norm. Defaults to ``None``.
            precision_tol (float, optional): threshold :math:`\\epsilon` that defines
                if Choi operator of the network is :math:`\\epsilon`-close to causality
                in the norm given by ``order``. Defaults to :math:`10^{-8}`.

        Returns:
            bool: Causal order condition.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if order is None and self._backend.__class__.__name__ == "TensorflowBackend":
            order = "euclidean"

        backend = self._backend

        dim_out = self.partition[-1]
        dim_in = self.partition[-2]

        trace_out = TraceOperation(dim_out, backend=backend).full()
        trace_in = TraceOperation(dim_in, backend=backend).full()

        if self._backend.__class__.__name__ == "PyTorchBackend":
            reduced = self._tensordot(self.full(), trace_out, dims=([-1], [0]))
            sub_comb = self._tensordot(reduced, trace_in, dims=([-1], [0]))
            expected = self._tensordot(sub_comb, trace_in / dim_in, dims=0)
        else:
            reduced = self._tensordot(self.full(), trace_out, axes=(-1, 0))
            sub_comb = self._tensordot(reduced, trace_in, axes=(-1, 0))
            expected = self._tensordot(sub_comb, trace_in / dim_in, axes=0)

        norm = self._backend.calculate_vector_norm(reduced - expected, order=order)

        if float(norm) > precision_tol:
            return False

        if len(self.partition) == 2:
            return True

        return QuantumComb(
            sub_comb, self.partition[:-2], pure=False, backend=self._backend
        ).is_causal(order, precision_tol)

    @classmethod
    def from_operator(
        cls, operator, partition=None, inverse=False, pure=False, backend=None
    ):  # pylint: disable=W0237
        comb = super().from_operator(operator, partition, None, pure, backend)
        if inverse:
            # Convert mathmetical convention of Choi operator to physical convention
            comb.partition = comb.partition[::-1]
            comb._tensor = comb._tensor.T  # pylint: disable=W0212
        return comb


class QuantumChannel(QuantumComb):
    """Stores a Quantum channel, which is a special case of quantum comb.

    A quantum channel is a quantum comb with only one input and one output.
    This class includes all quantum channels, unitary operators, and quantum states.

    To construct a `QuantumChannel` object, one can use the `QuantumNetwork.from_nparray` method.
    **Note**: if one try to construct a quantum network from a unitary operator or Choi operator,
    the first system will be the output.
    However, here we assume the first system is the input system.
    It is important to specify `inverse=True` when constructing by `QuantumNetwork.from_nparray`.

    Args:
        tensor (ndarray): the tensor representations of the quantum comb.
        partition (List[int] or Tuple[int], optional): partition of ``matrix``.
            If not provided and `system_input` is `None`, assume the input is a quantum state,
            whose input is a trivial system. If `system_input` is set to `True`,
            assume the input is an observable, whose output is a trivial system.
        system_input (List[bool] or Tuple[bool], optional): mask on the input system of the
            Choi operator. If ``None`` the default is ``(True,False)``.
            Defaults to ``None``.
        pure (bool, optional): ``True`` when ``tensor`` is a "pure" representation (e.g. a pure
            state, a unitary operator, etc.), ``False`` otherwise. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to the current backend.
            Defaults to ``None``.
    """

    def __init__(
        self,
        tensor,
        partition: Optional[Union[List[int], Tuple[int]]] = None,
        system_input: Optional[Union[List[bool], Tuple[bool]]] = None,
        pure: bool = False,
        backend=None,
    ):
        if isinstance(partition, int):
            partition = (partition,)

        if partition is not None:
            if len(partition) > 2:
                raise_error(
                    ValueError,
                    "A quantum channel should only contain one input system and one output system."
                    + "For general quantum networks, one should use the ``QuantumNetwork`` class.",
                )
            if len(partition) == 1:
                if system_input is None:  # Assume the input is a quantum state
                    partition = (1, partition[0])
                else:
                    if isinstance(system_input, bool):
                        system_input = (system_input,)

                    partition = (
                        (partition[0], 1) if system_input[0] else (1, partition[0])
                    )

        super().__init__(tensor, partition, pure=pure, backend=backend)

    def is_unital(
        self, order: Optional[Union[int, str]] = None, precision_tol: float = 1e-8
    ):
        """Returns bool indicating if the Choi operator :math:`\\mathcal{J}` is unital.

        A map is unital if it preserves the identity operator.
        Unitality is calculated as distance between the partial trace of :math:`\\mathcal{J}`
        and the Identity operator :math:`I`, with respect to a given norm.
        Default is the ``Hilbert-Schmidt`` norm (also known as ``Frobenius`` norm).

        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm
        <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.

        Args:
            order (str or int, optional): order of the norm. Defaults to ``None``.
            precision_tol (float, optional): threshold :math:`\\epsilon` that defines
                if Choi operator of the network is :math:`\\epsilon`-close to unitality
                in the norm given by ``order``. Defaults to :math:`10^{-8}`.

        Returns:
            bool: Unitality condition.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if order is None and self._backend.__class__.__name__ == "TensorflowBackend":
            order = "euclidean"

        backend = self._backend

        dim_out = self.partition[1]
        dim_in = self.partition[0]

        trace_out = TraceOperation(dim_out, backend=backend).full()
        trace_in = TraceOperation(dim_in, backend=backend).full()

        if self._backend.__class__.__name__ == "PyTorchBackend":
            reduced = self._tensordot(self.full(), trace_in, dims=([0], [0]))
            sub_comb = self._tensordot(
                reduced,
                trace_out,
                dims=([0], [0]),
            )
            expected = self._tensordot(trace_out / dim_out, sub_comb, dims=0)
        else:
            reduced = self._tensordot(self.full(), trace_in, axes=(0, 0))
            sub_comb = self._tensordot(reduced, trace_out, axes=(0, 0))
            expected = self._tensordot(trace_out / dim_out, sub_comb, axes=0)

        norm = self._backend.calculate_vector_norm((reduced - expected), order=order)
        if float(norm) > precision_tol:
            return False

        if len(self.partition) == 2:
            return True

        # Unital is defined for quantum channels only.
        # But we can extend it to quantum combs as follows:
        return QuantumChannel(  # pragma: no cover
            sub_comb, self.partition[2:], pure=False, backend=self._backend
        ).is_unital(order, precision_tol)

    def is_channel(
        self,
        order: Optional[Union[int, str]] = None,
        precision_tol_causal: float = 1e-8,
        precision_tol_psd: float = 1e-8,
    ):
        """Returns bool indicating if Choi operator :math:`\\mathcal{E}` is a channel.

        Args:
            order (int or str, optional): order of the norm used to calculate causality.
                Defaults to ``None``.
            precision_tol_causal (float, optional): threshold :math:`\\epsilon` that defines if
                Choi operator of the network is :math:`\\epsilon`-close to causality in the norm
                given by ``order``. Defaults to :math:`10^{-8}`.
            precision_tol_psd (float, optional): threshold value used to check if eigenvalues of
                the Choi operator :math:`\\mathcal{E}` are such that
                :math:`\\textup{eigenvalues}(\\mathcal{E}) >= \\textup{precision_tol_psd}`.
                Note that this parameter can be set to negative values.
                Defaults to :math:`0.0`.

        Returns:
            bool: Channel condition.
        """
        return self.is_causal(
            order, precision_tol_causal
        ) and self.is_positive_semidefinite(precision_tol_psd)

    def apply(self, state):
        """Apply the Choi operator :math:`\\mathcal{E}` to ``state`` :math:`\\varrho`.

        It is assumed that ``state`` :math:`\\varrho` is a density matrix.

        Args:
            state (ndarray): density matrix of a ``state``.

        Returns:
            ndarray: Resulting state :math:`\\mathcal{E}(\\varrho)`.
        """
        operator = self.copy().operator()
        conj = self._backend.np.conj

        if self.is_pure():
            return self._einsum("ij,lk,il", operator, conj(operator), state)

        return self._einsum("ijkl, jl", operator, state)


def link_product(
    subscripts: str,
    *operands: QuantumNetwork,
    backend=None,
    surpress_warning=False,
):
    """Link product between two quantum networks.

    The link product is not commutative. Here, we assume that
    :math:`A.\\textup{link_product}(B)` means "applying :math:`B` to :math:`A`".
    However, the ``link_product`` is associative, so we override the `@` operation
    in order to simplify notation.

    Args:
        subscripts (str, optional): Specifies the subscript for summation using
            the Einstein summation convention. For more details, please refer to
            `numpy.einsum <https://numpy.org/doc/stable/reference/generated/numpy.einsum.html>`_.
        operands (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum
            networks to be contracted.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to the current backend.
            Defaults to ``None``.
        surpress_warning (bool, optional): If ``True``, surpresses the warning
            regarding if the same index connects two input or two output
            systems. Defaults to ``False``.

    Returns:
        :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network resulting
            from the link product between two quantum networks.
    """

    if not isinstance(subscripts, str):
        raise_error(
            TypeError,
            f"subscripts must be type str, but it is type {type(subscripts)}.",
        )

    for i, operand in enumerate(operands):
        if not isinstance(operand, QuantumNetwork):
            raise_error(TypeError, f"The {i}-th operator is not a ``QuantumNetwork``.")

    if backend is None:  # pragma: no cover
        backend = operands[0]._backend  # pylint: disable=W0212

    tensors = [
        (
            backend.to_numpy(operand.full())
            if operand.is_pure()
            else backend.to_numpy(operand._tensor)  # pylint: disable=W0212
        )
        for operand in operands
    ]

    # keep track of the `partition` and `system_input` of the network
    _, contracrtion_list = np.einsum_path(
        subscripts, *tensors, optimize=False, einsum_call=True
    )

    inds, idx_rm, einsum_str, _, _ = contracrtion_list[0]
    input_str, results_index = einsum_str.split("->")
    inputs = input_str.split(",")

    # Warning if the same index connects two input or two output systems
    if not surpress_warning:
        for ind in idx_rm:
            found = 0
            for i, script in enumerate(inputs):
                index = script.find(ind)
                if index < 0:
                    continue
                found += 1
                if found > 1 and is_input == operands[inds[i]].system_input[index]:
                    warning(
                        f"Index {ind} connects two {'input' if is_input else 'output'} systems."
                    )
                is_input = operands[inds[i]].system_input[index]
                if found > 2:
                    warning(
                        f"Index {ind} appears multiple times in the input subscripts {input_str}."
                    )

    # set correct order of the `partition` and `system_input`
    partition = []
    system_input = []
    for ind in results_index:
        for i, script in enumerate(inputs):
            index = script.find(ind)
            if index < 0:
                continue

            partition.append(operands[inds[i]].partition[index])
            system_input.append(operands[inds[i]].system_input[index])

    new_tensor = np.einsum(subscripts, *tensors)

    return QuantumNetwork(new_tensor, partition, system_input, backend=backend)


class IdentityChannel(QuantumChannel):
    """The Identity channel with the given dimension.

    Args:
        dim (int): Dimension of the Identity operator.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to the current backend.
            Defaults to ``None``.
    """

    def __init__(self, dim: int, backend=None):

        identity = np.eye(dim, dtype=complex)
        identity = backend.cast(identity, dtype=identity.dtype)
        super().__init__(identity, [dim, dim], pure=True, backend=backend)


class TraceOperation(QuantumNetwork):
    """The Trace operator with the given dimension.

    Args:
        dim (int): Dimension of the Trace operator.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to the current backend.
            Defaults to ``None``.
    """

    def __init__(self, dim: int, backend=None):

        identity = np.eye(dim, dtype=complex)
        identity = backend.cast(identity, dtype=identity.dtype)
        super().__init__(identity, [dim], [True], pure=False, backend=backend)



"Module with the object that stores results from circuit execution using the HammingWeightBackend."

from typing import Optional, Union

import numpy as np
from scipy.special import binom

from qibo.backends import HammingWeightBackend
from qibo.config import raise_error
from qibo.result import MeasurementOutcomes, QuantumState


class HammingWeightResult(QuantumState, MeasurementOutcomes):
    """Object storing the results of a circuit execution with the
    :class:`qibo.backends.hamming_weight.HammingWeightBackend`.

    Args:
        state (ndarray): statevector with a fixed Hamming weight.
            The dimension of the state is :math:`d = \\binom{n}{k}`, where
            :math:`n` is the number of qubits and :math:`k` is the Hamming weight.
            The components of the state are ordered in lexicographical order.
        weight (int): Hamming weight of ``state``.
        nqubits (int): number of qubits of ``state``.
        measurements (list, optional): list of measurements gates :class:`qibo.gates.M`.
            Defaults to ``None``.
        nshots (int, optional): number of shots used for sampling the measurements.
            Defaults to :math:`1000`.
        engine (str, optional): engine to use in the execution of the
            :class:`qibo.backends.HammingWeightBackend`. It accepts ``"numpy"``, ``"numba"``,
            ``"cupy"``, and ``"cuquantum"``. If ``None``, defaults to the corresponding engine
            from the current backend. Defaults to ``None``.
    """

    def __init__(
        self,
        state,
        weight: int,
        nqubits: int,
        measurements=None,
        nshots: int = 1000,
        engine=None,
    ):  # pylint: disable=too-many-arguments

        backend = HammingWeightBackend(engine)
        QuantumState.__init__(self, state, backend)
        MeasurementOutcomes.__init__(self, measurements, backend, nshots=nshots)

        self.nqubits = nqubits
        self.weight = weight

        self._state = state

    def symbolic(self, decimals: int = 5, cutoff: float = 1e-10, max_terms: int = 20):
        """Dirac notation representation of the state in the computational basis.

        Args:
            decimals (int, optional): Number of decimals for the amplitudes.
                Defaults to :math:`5`.
            cutoff (float, optional): Amplitudes with absolute value smaller than the
                cutoff are ignored from the representation. Defaults to  :math:`1e-10`.
            max_terms (int, optional): Maximum number of terms to print. If the state
                contains more terms they will be ignored. Defaults to :math:`20`.

        Returns:
            str: String representing the state in the computational basis.
        """
        terms = self.backend.calculate_symbolic(
            self._state, self.nqubits, self.weight, decimals, cutoff, max_terms
        )
        return " + ".join(terms)

    def full_state(self):
        """Tensor representation of ``state`` in the entire computational basis.

        .. note::
            This method is inefficient in runtime and memory for a large number of qubits.
        """

        self.backend._dict_indexes = self.backend._get_lexicographical_order(
            self.nqubits, self.weight
        )

        state = self.backend.np.zeros(2**self.nqubits, dtype=self.backend.np.complex128)
        state = self.backend.cast(state, dtype=state.dtype)
        indices = list(self.backend._dict_indexes.values())
        indices.sort()
        indices = np.asarray(indices)
        state[list(indices[:, 1])] = self._state[list(indices[:, 0])]

        return state

    def probabilities(self, qubits: Optional[Union[list, set]] = None):
        """Calculate the probabilities of the measured qubits.

        If the number of shots is :math:`0` or no measurements were performed,
        the probabilities are calculated from the statevector.
        Otherwise, the probabilities are calculated from the samples.

        Args:
            qubits (list or set, optional): Set of qubits that are measured.
                If ``None``, ``qubits`` equates the total number of qubits.
                Defauts to ``None``.
        Returns:
            ndarray: Probabilities over the input qubits.
        """

        if self.nshots is None or len(self.measurements) == 0:
            return self._exact_probabilities(qubits)

        return self._probabilities_from_samples(qubits)

    def _exact_probabilities(self, qubits: Optional[Union[list, set]] = None):
        """Calculate measurement probabilities by tracing out qubits.

        Args:
            qubits (list or set, optional): Set of qubits that are measured.
                If ``None``, ``qubits`` equates the total number of qubits.
                Defauts to ``None``.

        Returns:
            ndarray: Probabilities over the input qubits.
        """

        if qubits is None:
            qubits = tuple(range(self.nqubits))

        return self.backend.calculate_probabilities(
            self._state, qubits, self.weight, self.nqubits
        )

    def _probabilities_from_samples(self, qubits: Optional[Union[list, set]] = None):
        """Calculate the probabilities as ``frequencies / nshots``.

        Args:
            qubits (list or set, optional): Set of qubits that are measured.
                If ``None``, ``qubits`` equates the total number of qubits.
                Defauts to ``None``.

        Returns:
            ndarray: Array containing the probabilities of the measured qubits.
        """
        if qubits is None:
            qubits = self.measurement_gate.qubits
        else:
            if not set(qubits).issubset(self.measurement_gate.qubits):
                raise_error(
                    RuntimeError,
                    f"Asking probabilities for qubits {qubits}, "
                    + f"but only qubits {self.measurement_gate.qubits} were measured.",
                )
        qubits = [self.measurement_gate.qubits.index(q) for q in qubits]

        nqubits = len(self.measurement_gate.qubits)
        probs = [0 for _ in range(2**nqubits)]
        for state, freq in self.frequencies(binary=False).items():
            probs[state] = freq / self.nshots
        rtype = self.backend.np.real(probs).dtype
        probs = self.backend.cast(probs, dtype=rtype)
        self._probs = probs

        if nqubits != self.nqubits:
            self.backend._dict_indexes = None

        return self.backend.calculate_full_probabilities(
            self.backend.np.sqrt(probs), qubits, nqubits
        )

    def samples(self, binary: bool = True, registers: bool = False):
        """Returns raw measurement samples.

        Args:
            binary (bool, optional): Return samples in binary or decimal form.
                Defaults to ``True``.
            registers (bool, optional): Group samples according to registers.
                Defaults to ``False``.

        Returns:
            If ``binary`` is ``True``
                samples are returned in binary form as a tensor
                of shape ``(nshots, n_measured_qubits)``.
            If ``binary`` is ``False``
                samples are returned in decimal form as a tensor
                of shape ``(nshots,)``.
            If ``registers`` is ``True``
                samples are returned in a ``dict`` where the keys are the register
                names and the values are the samples tensors for each register.
            If ``registers`` is ``False``
                a single tensor is returned which contains samples from all the
                measured qubits, independently of their registers.
        """

        if len(self.measurements) == 0:
            raise_error(
                RuntimeError,
                "No measurements were performed. Cannot return samples.",
            )

        self._probs = self._exact_probabilities()
        return super().samples(binary=binary, registers=registers)

    def frequencies(self, binary: bool = True, registers: bool = False):
        """Return the frequencies of measured samples.

        Args:
            binary (bool, optional): If ``True``, returns frequency keys in binary form.
                If ``False``, returns them in decimal form. Defaults to ``True``.
            registers (bool, optional): Group frequencies according to registers.
                Defaults to ``False``.

        Returns:
            A :class:`collections.Counter` where the keys are the observed values
            and the values the corresponding frequencies, that is the number
            of times each measured value/bitstring appears.

            If ``binary`` is ``True``
                the keys of the :class:`collections.Counter` are in binary form,
                as strings of :math:`0` and :math`1`.
            If ``binary`` is ``False``
                the keys of the :class:`collections.Counter` are integers.
            If ``registers`` is ``True``
                a `dict` of :class:`collections.Counter` is returned where keys are
                the name of each register.
            If ``registers`` is ``False``
                a single :class:`collections.Counter` is returned which contains samples
                from all the measured qubits, independently of their registers.
        """
        if len(self.measurements) == 0:
            raise_error(
                RuntimeError,
                "No measurements were performed. Cannot return frequencies.",
            )

        if not self.has_samples():
            self._probs = self._exact_probabilities(self._measurement_gate.qubits)

        return super().frequencies(binary=binary, registers=registers)



from abc import ABC, abstractmethod
from typing import Tuple

import networkx as nx

from qibo.models import Circuit


class Placer(ABC):
    """Maps logical qubits to physical qubits."""

    @abstractmethod
    def __init__(self, connectivity: nx.Graph, *args):
        """Initializes the placer.

        Args:
            connectivity (nx.Graph): Hardware topology.
        """

    @abstractmethod
    def __call__(self, circuit: Circuit, *args):
        """Find initial qubit mapping.

        Method works in-place.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be placed.
        """


class Router(ABC):
    """Makes the circuit executable on the given topology."""

    @abstractmethod
    def __init__(self, connectivity: nx.Graph, *args):
        """Initializes the router.

        Args:
            connectivity (nx.Graph): Hardware topology.
        """

    @abstractmethod
    def __call__(self, circuit: Circuit, *args) -> Tuple[Circuit, dict]:
        """Match circuit to hardware connectivity.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be routed.

        Returns:
            (:class:`qibo.models.circuit.Circuit`, dict): Routed circuit and final {logical: physical} qubit mapping.
        """


class Optimizer(ABC):
    """Reduces the number of gates in the circuit."""

    @abstractmethod
    def __call__(self, circuit: Circuit, *args) -> Circuit:
        """Optimize transpiled circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be optimized.

        Returns:
            (:class:`qibo.models.circuit.Circuit`): Optimized circuit.
        """



from qibo.models import hep, tsp
from qibo.models.circuit import Circuit
from qibo.models.encodings import (
    binary_encoder,
    comp_basis_encoder,
    dicke_state,
    entangling_layer,
    ghz_state,
    graph_state,
    hamming_weight_encoder,
    permutation_synthesis,
    phase_encoder,
    sparse_encoder,
    unary_encoder,
    unary_encoder_random_gaussian,
)
from qibo.models.error_mitigation import CDR, ICS, ZNE, vnCDR
from qibo.models.evolution import AdiabaticEvolution, StateEvolution
from qibo.models.grover import Grover
from qibo.models.qft import QFT
from qibo.models.variational import AAVQE, FALQON, QAOA, VQE



"""Module defining the Hamming-weight-preserving backend."""

from typing import List, Union

import numpy as np
from scipy.special import binom

from qibo import gates
from qibo.config import raise_error


def apply_gate(self, gate, state, nqubits: int, weight: int):
    """Apply ``gate`` to ``state``.

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): gate to apply to ``state``.
        state (ndarray): state to apply ``gate`` to.
        nqubits (int): total number of qubits in ``state``.
        weight (int): fixed Hamming weight of ``state``.

    Returns:
        ndarray: ``state`` after the action of ``gate``.
    """
    if isinstance(gate, gates.M):
        return gate.apply_hamming_weight(self, state, nqubits, weight)

    if isinstance(gate, gates.CCZ):
        # CCZ has a custom apply method because currently it is the only
        # 3-qubit gate that is also Hamming-weight-preserving
        # and this custom method is faster than the n-qubit method
        return self._apply_gate_CCZ(gate, state, nqubits, weight)

    if len(gate.target_qubits) == 1:
        return self._apply_gate_single_qubit(gate, state, nqubits, weight)

    if len(gate.target_qubits) == 2:
        return self._apply_gate_two_qubit(gate, state, nqubits, weight)

    return self._apply_gate_n_qubit(gate, state, nqubits, weight)


def execute_circuit(self, circuit, weight: int, initial_state=None, nshots: int = 1000):
    """Execute ``circuit`` by applying the queue of gates to the ``initial_state``.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Hamming-weight-preserving circuit
            to be executed.
        weight (int): fixed Hamming weight of the ``initial_state``.
        initial_state (ndarray, optional): initial state that ``circuit`` acts on.
            If ``None``, defaults to :math:`\\ket{0^{n-k} \\, 1^{k}}`,
            with :math:`n` being the total number of qubits in the circuit,
            and :math:`k` being the Hamming ``weight``. Defaults to ``None``.
        nshots (int, optional): total number of shots to simulate when ``circuit``
            contains measurement gates (:class:`qibo.gates.M`). Defaults to :math:`1000`.

    Returns:
        :class:`qibo.quantum_info.hamming_weight.HammingWeightResult`: Object
        containing the results of circuit execution of a Hamming-weight-preserving
        circuit.
    """
    from qibo.quantum_info.hamming_weight import (  # pylint: disable=import-outside-toplevel
        HammingWeightResult,
    )

    if circuit.density_matrix:
        raise_error(RuntimeError, "Density matrix simulation is not supported.")

    for gate in circuit.queue:
        if isinstance(gate, gates.Channel):
            raise_error(RuntimeError, "Circuit must not contain channels.")
        if not gate.hamming_weight and not isinstance(gate, gates.M):
            raise_error(
                RuntimeError,
                "Circuit contains non-Hammming weight preserving  gates.",
            )

    nqubits = circuit.nqubits

    self._dict_indexes = self._get_lexicographical_order(nqubits, weight)

    try:
        if initial_state is None:
            n_choose_k = int(binom(nqubits, weight))
            initial_state = self.np.zeros(n_choose_k)
            initial_state[0] = 1
            initial_state = self.cast(initial_state, dtype=self.dtype)

        state = initial_state
        for gate in circuit.queue:
            state = self.apply_gate(gate, state, nqubits, weight)

        result = HammingWeightResult(
            state,
            weight,
            nqubits,
            measurements=circuit.measurements,
            nshots=nshots,
            engine=self.platform,
        )

        return result

    except self.oom_error:  # pragma: no cover
        raise_error(
            RuntimeError,
            f"State does not fit in {self.device} memory."
            "Please switch the execution device to a "
            "different one using ``qibo.set_device``.",
        )


def _gray_code(self, initial_string):
    """Return all bitstrings of a fixed Hamming weight.

    Uses the ``ehrlich_algorithm`` with an ``initial_string``.

    Args:
        initial_string (ndarray): Array of bits representing the input
            of the Ehrlich algorithm.

    Returns:
        ndarray: All bitstrings with the same Hamming weight as ``initial_string``.
    """
    from qibo.models.encodings import _ehrlich_algorithm  # pylint: disable=C0415

    strings = _ehrlich_algorithm(initial_string, return_indices=False)
    strings = [list(string) for string in strings]
    strings = self.np.asarray(strings, dtype=int)

    return strings


def _get_cached_strings(
    self, nqubits: int, weight: int, ncontrols: int = 0, two_qubit_gate: bool = True
):
    """Generate list of strings necessary for the custom ``apply_gate`` method.

    Given the total number of qubits ``nqubits``, the Hamming ``weight`` to be
    preserved, and the number of controls ``ncontrols`` in the gate, returns a
    sequence of bitstrings for the custom ``apply_gate`` method.

    The sequence is generated for the first gate with a unique combination of ``nqubits``,
    ``weight`` and ``ncontrols``, and then cached to be resued for similar gates.

    Args:
        nqubits (int): total number of qubits in the quantum system.
        weight (int): Hamming weight of the state that gates are acting on.
        ncontrols (int, optional): number of controls in the gate that is being
            applied to the state. Defaults to :math:`0`.
        two_qubit_gate (bool, optional): if ``True``, generate strings assuming the
            gate is a two-qubit gate. If ``False``, it assumes the gate is a single-qubit
            gate. Defaults to ``True``.

    Returns:
        ndarray or list: ndarray of bitstrings for two-qubit gates or a list of two ndarrays
        of bitstrings for single-qubit gates.
    """
    if two_qubit_gate:
        initial_string = self.np.array(
            [1] * (weight - 1 - ncontrols)
            + [0] * ((nqubits - 2 - ncontrols) - (weight - 1 - ncontrols))
        )
        strings = self._gray_code(initial_string)
    else:
        initial_string = self.np.array(
            [1] * (weight - ncontrols)
            + [0] * ((nqubits - 1 - ncontrols) - (weight - ncontrols))
        )
        strings_0 = self._gray_code(initial_string)

        initial_string = self.np.array(
            [1] * (weight - 1 - ncontrols)
            + [0] * ((nqubits - 1 - ncontrols) - max(0, (weight - 1 - ncontrols)))
        )
        strings_1 = self._gray_code(initial_string)

        strings = [strings_0, strings_1]

    return strings


def _get_lexicographical_order(self, nqubits, weight):
    """Sort bistrings generated from ``self._get_cached_strings`` in lexicographical order.

    Bitstrings are sorted in lexicographical (ascending) order.
    Moreover, they are also converted to integers and both representations are cached,
    creating a map between the full statevector representation and the condensed
    Hamming-weight-preserving representation.

    Args:
        nqubits (int): total number of qubits.
        weight (int): Hamming weight of the state.

    Returns:
        dict: Dictionary with the cached bitstrings and the aforementioned map.
    """
    n_choose_k = int(binom(nqubits, weight))
    indexes = list(range(n_choose_k))

    lexicographical_order = self._get_cached_strings(nqubits + 2, weight + 1)
    lexicographical_order = [
        "".join(item.astype(str)) for item in lexicographical_order
    ]
    lexicographical_order.sort()
    lexicographical_order_int = [int(item, base=2) for item in lexicographical_order]
    _dict_indexes = dict(
        zip(lexicographical_order, zip(indexes, lexicographical_order_int))
    )

    return _dict_indexes


def _get_single_qubit_matrix(self, gate):
    """Return non-zero elements of the matrix representation of
    Hamming-weight-preserving single-qubit gates.

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Hamming-weight-preserving
            single-qubit gate.

    Returns:
        tuple(complex, complex): Non-zero elements of a Hamming-weight-preserving ``gate``.
    """
    matrix = gate.matrix(backend=self)

    if gate.name in ["cz", "crz", "cu1"]:
        matrix = matrix[2:, 2:]

    return self.np.diag(matrix)


def _apply_gate_single_qubit(self, gate, state, nqubits, weight):
    """Custom ``apply_gate`` method for Hamming-weight-preserving single-qubit gates.

    Instead of relying on matrix multiplication, this method applies
    Hamming-weight-preserving single-qubit gates by directly multiplying
    the amplitudes of interest elementwise by the necessary phase(s).

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Hamming-weight-preserving
            single-qubit gate to be applied to ``state``
        state (ndarray): state to suffer the action of ``gate``.
        nqubits (int): total number of qubits in the circuit.
        weight (int): Hamming weight of ``state``.

    Returns:
        ndarray: ``state`` after the action of ``gate``.
    """
    qubits = list(gate.target_qubits)
    controls = list(gate.control_qubits)
    ncontrols = len(controls)
    other_qubits = list(set(range(nqubits)) ^ set(qubits + controls))

    key_0, key_1 = f"{ncontrols}_0", f"{ncontrols}_1"
    if (
        key_0 not in self._dict_cached_strings_one
        or key_1 not in self._dict_cached_strings_one
    ):
        strings_0, strings_1 = self._get_cached_strings(
            nqubits, weight, ncontrols, False
        )
        self._dict_cached_strings_one[key_0] = strings_0
        self._dict_cached_strings_one[key_1] = strings_1
    else:
        strings_0 = self._dict_cached_strings_one[key_0]
        strings_1 = self._dict_cached_strings_one[key_1]

    matrix_00, matrix_11 = self._get_single_qubit_matrix(gate)

    indexes_one = self.np.zeros((len(strings_1), nqubits), dtype=str)

    if matrix_00 != 1.0 and nqubits - weight > 0:
        indexes_zero = self.np.zeros((len(strings_0), nqubits), dtype=str)
        indexes_zero[:, other_qubits] = strings_0
        indexes_zero[:, qubits] = ["0"]
        if len(controls) > 0:
            indexes_zero[:, controls] = "1"
        indexes_zero = self.np.array(
            [self._dict_indexes["".join(elem)][0] for elem in indexes_zero]
        )

        state[indexes_zero] *= matrix_00

    if matrix_11 != 1.0 and weight - ncontrols > 0:
        indexes_one[:, other_qubits] = strings_1
        if len(controls) > 0:
            indexes_one[:, controls] = "1"

        indexes_one[:, qubits] = ["1"]
        indexes_one = self.np.array(
            [self._dict_indexes["".join(elem)][0] for elem in indexes_one]
        )

        state[indexes_one] *= matrix_11

    return state


def _update_amplitudes(
    self,
    state,
    qubits: List[int],
    controls: List[int],
    other_qubits: List[int],
    weight: int,
    matrix_element: Union[complex, float],
    bitlist: List[str],
    shift: int,
):
    """Update in-place the amplitudes changed by a two-qubit Hamming-weight-preserving gate.

    Args:
        state (ndarray): state that the two-qubit gate acts on.
        qubits (list): target qubits of the gate.
        controls (list): control qubits of the gate.
        other_qubits (list): remaining qubits in the circuit.
        weight (int): Hamming weight of ``state``.
        matrix_element (complex or float): non-zero element of the gate that
            multiplies the amplitudes of ``state``.
        bitlist (list): If the amplitude being updated is associated to the computational basis
            state :math:`\\ket{\\dots00\\dots}`, then ``bitlist=["0", "0"]``.
            If the amplitude is associated to :math:`\\ket{\\dots11\\dots}`,
            then ``bitlist=["1", "1"]``.
        shift (int): shift in the Hamming weight necessary to generate bitstrings.
            If the amplitude being updated is associated to the computational basis
            state :math:`\\ket{\\dots00\\dots}`, then ``shift`` is :math:`1`.
            If the amplitude is associated to :math:`\\ket{\\dots11\\dots}`,
            then ``shift`` is :math:`-1`.

    Returns:
        ndarray: ``state`` after the action of two-qubit Hamming-weight-preserving gate.
    """
    ncontrols = len(controls)
    nqubits = len(qubits) + ncontrols + len(other_qubits)

    strings = self._get_cached_strings(nqubits, weight + shift, ncontrols)
    indexes_in = self.np.zeros((len(strings), nqubits), dtype=str)
    indexes_in[:, other_qubits] = strings
    if ncontrols > 0:
        indexes_in[:, controls] = "1"
    indexes_in[:, qubits] = bitlist
    indexes_in = self.np.array(
        [self._dict_indexes["".join(elem)][0] for elem in indexes_in]
    )
    state[indexes_in] *= matrix_element

    return state


def _apply_gate_two_qubit(self, gate, state, nqubits, weight):
    """Custom ``apply_gate`` method for Hamming-weight-preserving two-qubit gates.

    Instead of relying on matrix multiplication, this method applies
    Hamming-weight-preserving two-qubit gates by directly multiplying
    the amplitudes of interest elementwise by the necessary phase(s).

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Hamming-weight-preserving
            two-qubit gate to be applied to ``state``
        state (ndarray): state to suffer the action of ``gate``.
        nqubits (int): total number of qubits in the circuit.
        weight (int): Hamming weight of ``state``.

    Returns:
        ndarray: ``state`` after the action of ``gate``.
    """
    qubits = list(gate.target_qubits)
    controls = list(gate.control_qubits)
    ncontrols = len(controls)
    other_qubits = list(set(list(range(nqubits))) ^ set(qubits + controls))

    key = f"{ncontrols}"
    if key not in self._dict_cached_strings_two:
        self._dict_cached_strings_two[key] = self._get_cached_strings(
            nqubits, weight, ncontrols
        )

    strings = self._dict_cached_strings_two[key]

    matrix = gate.matrix(backend=self)
    matrix_0000, matrix_1111 = matrix[0, 0], matrix[3, 3]
    matrix_0101, matrix_0110 = matrix[1, 1], matrix[1, 2]
    matrix_1001, matrix_1010 = matrix[2, 1], matrix[2, 2]

    if weight - ncontrols > 0 and weight not in [0, nqubits]:
        indexes_in = self.np.zeros((len(strings), nqubits), dtype=str)
        indexes_in[:, other_qubits] = strings
        if len(controls) > 0:
            indexes_in[:, controls] = "1"
        indexes_in[:, qubits] = ["1", "0"]
        indexes_out = self.np.copy(indexes_in)
        indexes_out[:, qubits] = ["0", "1"]
        indexes_in = self.np.array(
            [self._dict_indexes["".join(elem)][0] for elem in indexes_in]
        )
        indexes_out = self.np.array(
            [self._dict_indexes["".join(elem)][0] for elem in indexes_out]
        )

        old_in, old_out = state[indexes_in], state[indexes_out]
        new_amplitudes_in = matrix_1010 * old_in + matrix_1001 * old_out
        new_amplitudes_out = matrix_0101 * old_out + matrix_0110 * old_in

        state[indexes_in] = new_amplitudes_in
        state[indexes_out] = new_amplitudes_out

    if (
        weight - ncontrols >= 0
        and nqubits - weight > 1
        and (matrix_0000.real != 1 or abs(matrix_0000.imag) > 0)
    ):
        # update the |...00...> amplitudes if necessary
        state = self._update_amplitudes(
            state,
            qubits,
            controls,
            other_qubits,
            weight,
            matrix_0000,
            ["0", "0"],
            shift=1,
        )

    if weight - ncontrols > 1 and (matrix_1111.real != 1 or abs(matrix_1111.imag) > 0):
        # update the |...11...> amplitudes if necessary
        state = self._update_amplitudes(
            state,
            qubits,
            controls,
            other_qubits,
            weight,
            matrix_1111,
            ["1", "1"],
            shift=-1,
        )

    return state


def _apply_gate_CCZ(self, gate, state, nqubits: int, weight: int):
    """Custom ``apply_gate`` method for the :class:`qibo.gates.CCZ` gate.

    Args:
        gate (:class:`qibo.gates.CCZ`): :math:`2`-controlled :math:`Z` gate
            to be applied to ``state``.
        state (ndarray): state to suffer the action of ``gate``.
        nqubits (int): total number of qubits in the circuit.
        weight (int): Hamming-weight of ``state``.

    Returns:
        ndarray: ``state`` after the action of the :class:`qibo.gates.CCZ` gate.
    """
    qubits = list(gate.qubits)
    gate_qubits = len(qubits)

    self._dict_indexes = self._get_lexicographical_order(nqubits, weight)

    strings = list(self._dict_indexes.keys())

    for j in range(len(strings)):
        gate_string = [strings[j][q] for q in qubits]
        if gate_string.count("1") == gate_qubits:
            state[j] *= -1

    return state


def _apply_gate_n_qubit(self, gate, state, nqubits, weight):
    """Custom ``apply_gate`` method for Hamming-weight-preserving n-qubit gates.

    This method performs matrix multiplication directly in the subspace with Hamming-weight
    ``weight`` without the need to calculate the full matrix representation of the gate.

    .. note::
        The attribute ``gate.hamming_weight`` must be manually set to ``True`` for this method to work.

    Args:
        gate (:class:`qibo.gates.abstract.Gate`): Hamming-weight-preserving
            :math:`n`-qubit gate to be applied to ``state``
        state (ndarray): state to suffer the action of ``gate``.
        nqubits (int): total number of qubits in the circuit.
        weight (int): Hamming weight of ``state``.

    Returns:
        ndarray: ``state`` after the action of ``gate``.
    """
    gate_matrix = gate.matrix(backend=self)
    qubits = list(gate.target_qubits)
    gate_qubits = len(qubits)
    if 2 ** (gate_qubits) != gate_matrix.shape[0]:
        qubits = list(gate.qubits)
        gate_qubits = len(qubits)
        controls = []
        ncontrols = 0
    else:
        controls = list(gate.control_qubits)
        ncontrols = len(controls)

    other_qubits = list(set(qubits + controls) ^ set(range(nqubits)))
    map_ = qubits + controls + other_qubits
    gate_matrix = gate.matrix(backend=self)

    self._dict_indexes = self._get_lexicographical_order(nqubits, weight)

    strings = self.np.array(list(self._dict_indexes.keys()))
    indexes = self.np.array([index[1] for index in self._dict_indexes.values()])
    dim = len(indexes)

    matrix = self.np.zeros((dim, dim))
    matrix = self.cast(matrix, dtype=self.dtype)
    mod_condition = indexes[:, None] % 2 ** (
        nqubits - gate_qubits - ncontrols
    ) == indexes[None, :] % 2 ** (nqubits - gate_qubits - ncontrols)

    control_substrings = self.np.array(
        [s[gate_qubits : gate_qubits + ncontrols].count("1") for s in strings]
    )
    control_condition = control_substrings[:, None] == ncontrols

    row_indices = indexes[:, None] // 2 ** (nqubits - gate_qubits)
    col_indices = indexes[None, :] // 2 ** (nqubits - gate_qubits)

    matrix[mod_condition & control_condition] = gate_matrix[row_indices, col_indices][
        mod_condition & control_condition
    ]

    diagonal_indices = self.np.eye(dim, dtype=bool)
    matrix[mod_condition & ~control_condition & diagonal_indices] = 1

    new_matrix = self.np.zeros((dim, dim))
    new_matrix = self.cast(new_matrix, dtype=self.dtype)

    strings_array = self.np.array([list(s) for s in strings])
    reordered_strings_array = strings_array[:, map_]

    reordered_strings = self.np.array(["".join(s) for s in reordered_strings_array])
    reordered_indexes = [
        self.np.where(strings == new_string_i)[0][0]
        for new_string_i in reordered_strings
    ]
    new_matrix = matrix[reordered_indexes][:, reordered_indexes]

    new_matrix = self.cast(new_matrix)
    state = new_matrix @ state

    return state


def calculate_symbolic(
    self, state, nqubits, weight, decimals=5, cutoff=1e-10, max_terms=20
):
    """Dirac notation representation of the state in the computational basis.

    Args:
        state (ndarray): state to suffer the action of ``gate``.
        nqubits (int): total number of qubits in the circuit.
        weight (int): Hamming-weight of ``state``.
        decimals (int, optional): Number of decimals for the amplitudes.
            Defaults to :math:`5`.
        cutoff (float, optional): Amplitudes with absolute value smaller than the
            cutoff are ignored from the representation. Defaults to  :math:`1e-10`.
        max_terms (int, optional): Maximum number of terms to print. If the state
            contains more terms they will be ignored. Defaults to :math:`20`.

    Returns:
        str: String representing the state in the computational basis.
    """
    state = self.to_numpy(state)
    terms = []

    self._dict_indexes = self._get_lexicographical_order(nqubits, weight)

    strings = list(self._dict_indexes.keys())
    for i in self.np.nonzero(state)[0]:
        i = int(i)
        b = strings[i]
        if self.np.abs(state[i]) >= cutoff:
            x = self.np.round(state[i], decimals)
            terms.append(f"{x}|{b}>")
        if len(terms) >= max_terms:
            terms.append("...")
            return terms
    return terms


def calculate_probabilities(self, state, qubits, weight, nqubits):
    """Calculate the probabilities of the measured qubits from the statevector.

    Args:
        state (ndarray): state to suffer the action of ``gate``.
        qubits (list or set, optional): Set of qubits that are measured.
        weight (int): Hamming-weight of ``state``.
        nqubits (int): total number of qubits in the circuit.

    Returns:
        ndarray: Probabilities over the input qubits.
    """
    rtype = self.np.real(state).dtype

    self._dict_indexes = self._get_lexicographical_order(nqubits, weight)

    strings = list(self._dict_indexes.keys())
    indexes = [index[0] for index in self._dict_indexes.values()]

    measured_strings = {}
    for string, index in zip(strings, indexes):
        measured_string = "".join(string[q] for q in qubits)
        measured_strings[measured_string] = measured_strings.get(
            measured_string, 0
        ) + self.np.real(self.np.abs(state[index]) ** 2)

    strings = list(measured_strings.keys())
    indexes = [int(string, 2) for string in strings]
    probs = self.np.zeros(2 ** len(qubits), dtype=rtype)
    for index, string in zip(indexes, strings):
        probs[index] = measured_strings[string]

    probs = self.cast(probs, dtype=rtype)
    return probs


def collapse_state(self, state, qubits, shot, weight, nqubits, normalize=True):
    """Collapse state vector according to measurement shot.

    Args:
        state (ndarray): state to suffer the action of ``gate``.
        qubits (list or set, optional): Set of qubits that are measured.
        shot (list): Decimal value of the bitstring measured.
        weight (int): Hamming-weight of ``state``.
        nqubits (int): total number of qubits in the circuit.

    Returns:
        ndarray: collapsed ``state``.
    """
    state = self.cast(state, dtype=state.dtype)

    self._dict_indexes = self._get_lexicographical_order(nqubits, weight)

    strings = list(self._dict_indexes.keys())
    indexes = [index[0] for index in self._dict_indexes.values()]
    binshot = self.samples_to_binary(shot, len(qubits))[0]
    binshot = "".join(str(s) for s in binshot)
    for string, index in zip(strings, indexes):
        if "".join(string[q] for q in qubits) != binshot:
            state[index] = 0

    if normalize:
        norm = self.np.sqrt(self.np.sum(self.np.abs(state) ** 2))
        state = state / norm

    return state
